#/********************************************************************
# onvif-gui/onvif_gui/panels/settings/proxy.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QMessageBox, QLineEdit, QPushButton, \
    QGridLayout, QWidget, QLabel, QMessageBox, QRadioButton, \
    QGroupBox, QCheckBox, QComboBox, QSpinBox
from PyQt6.QtCore import Qt
from onvif_gui.components import DirectorySelector
from onvif_gui.enums import ProxyType
from loguru import logger
import libonvif as onvif
import ipaddress
import os
import sys
import html

class ProxyOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw

        self.proxyTypeKey = "settings/proxyType"
        self.proxyRemoteKey = "settings/proxyRemote"
        self.autoDownloadKey = "settings/autoDownloadMTX"
        self.listenKey = "settings/alarmsFromServer"
        self.broadcastInterfaceKey = "settings/proxyBroadcastInterface"
        self.broadcastEnableKey = "setings/enableBroadcasting"
        self.broadcastAllInterfacesKey = "settings/broadcastAllInterfaces"
        self.timeoutKey = "settings/clientSocketTimeout"
        self.httpKey = "settings/enableHttpServer"

        self.if_addrs = []
        session = onvif.Session()
        session.getActiveInterfaces()
        i = 0
        while len(session.active_interface(i)) > 0 and i < 16:
            if session.active_interface(i) != "172.17.0.1":
                self.if_addrs.append(session.active_interface(i))
            i += 1

        self.grpProxyType = QGroupBox("Select Proxy Type")
        self.radStandAlone = QRadioButton("Stand Alone", self.grpProxyType)
        self.radServer = QRadioButton("Server", self.grpProxyType)
        self.radClient = QRadioButton("Client", self.grpProxyType)

        self.lblServer = QLabel("\n\n")
        self.lblConnect = QLabel("Connect url for clients")
        self.lblConnect.setEnabled(False)

        self.chkAutoDownload = QCheckBox("Auto Download Media MTX")
        self.chkAutoDownload.setEnabled(False)
        self.chkAutoDownload.setChecked(int(mw.settings.value(self.autoDownloadKey, 1)))
        self.chkAutoDownload.stateChanged.connect(self.chkAutoDownloadChecked)

        self.txtDirectoryMTX = DirectorySelector(mw, "MTXDir", "Dir", self.getProxyServerDir())
        self.txtDirectoryMTX.setEnabled(not self.chkAutoDownload.isChecked())

        self.cmbProxyLog = QComboBox()
        self.cmbProxyLog.addItems(["error", "warn", "info", "debug"])

        self.grpMediaMTX = QGroupBox("Media MTX Settings")
        lytMediaMTX = QGridLayout(self.grpMediaMTX)
        lytMediaMTX.addWidget(self.chkAutoDownload,       0, 0, 1, 2)
        lytMediaMTX.addWidget(self.txtDirectoryMTX,       1, 0, 1, 2)
        lytMediaMTX.addWidget(QLabel("Log Level"),        2, 0, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytMediaMTX.addWidget(self.cmbProxyLog,           2, 1, 1, 1)

        self.cmbInterfaces = QComboBox()
        self.lblInterfaces = QLabel("Network")

        # Windows can only broadcast on the network interface with highest priority
        # this needs further research
        if sys.platform == "win32":
            self.cmbInterfaces.addItem(session.primary_network_interface())
        else:
            for if_addr in self.if_addrs:
                self.cmbInterfaces.addItem(if_addr)
        
        if len(self.if_addrs):
            self.cmbInterfaces.setCurrentText(self.mw.settings.value(self.broadcastInterfaceKey, self.if_addrs[0]))
        else:
            logger.error("No active network interface was found")
        self.cmbInterfaces.currentTextChanged.connect(self.cmbInterfacesChanged)
        
        self.grpAlarmBroadcast = QGroupBox("Alarm Broadcasting")
        self.grpAlarmBroadcast.setCheckable(True)
        self.grpAlarmBroadcast.setChecked(int(self.mw.settings.value(self.broadcastEnableKey, 1)))
        self.grpAlarmBroadcast.clicked.connect(self.grpAlarmBroadcastClicked)
        lytAlarmBroadcast = QGridLayout(self.grpAlarmBroadcast)
        lytAlarmBroadcast.addWidget(self.lblInterfaces,              1, 0, 1, 1)
        lytAlarmBroadcast.addWidget(self.cmbInterfaces,              1, 1, 1, 1)

        self.proxyRemote = self.mw.settings.value(self.proxyRemoteKey)

        self.txtRemote = QLineEdit()
        self.txtRemote.setText(self.proxyRemote)
        self.txtRemote.textEdited.connect(self.txtRemoteEdited)
        self.txtRemote.setEnabled(False)
        self.lblRemote = QLabel("Connect url from server")
        self.lblRemote.setEnabled(False)

        self.btnUpdate = QPushButton("Update")
        self.btnUpdate.clicked.connect(self.btnUpdateClicked)
        self.btnUpdate.setEnabled(False)

        self.chkListen = QCheckBox("Get alarm events from server")
        self.chkListen.setChecked(int(mw.settings.value(self.listenKey, 1)))
        self.chkListen.stateChanged.connect(self.chkListenChecked)
        self.chkListen.setEnabled(False)

        self.spnTimeout = QSpinBox()
        self.spnTimeout.setValue(int(self.mw.settings.value(self.timeoutKey, 5)))
        self.spnTimeout.valueChanged.connect(self.spnTimeoutChanged)

        self.lblTimeout = QLabel("Client socket timeout (seconds)")
        pnlTimeout = QWidget()
        lytTimeout = QGridLayout(pnlTimeout)
        lytTimeout.addWidget(self.lblTimeout,  0, 0, 1, 1)
        lytTimeout.addWidget(self.spnTimeout,  0, 1, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytTimeout.setContentsMargins(0, 0, 0, 0)

        self.chkEnableHttpServer = QCheckBox("Enable HTTP server")
        self.chkEnableHttpServer.setChecked(int(mw.settings.value(self.httpKey, 0)))
        self.chkEnableHttpServer.stateChanged.connect(self.chkEnableHttpServerChecked)

        self.proxyType = None
        match self.mw.settings.value(self.proxyTypeKey, ProxyType.STAND_ALONE):
            case ProxyType.STAND_ALONE:
                self.radStandAlone.setChecked(True)
                self.radStandAloneToggled(True)
            case ProxyType.SERVER:
                self.radServer.setChecked(True)
                self.radServerToggled(True)
            case ProxyType.CLIENT:
                self.radClient.setChecked(True)
                self.radClientToggled(True)

        self.radStandAlone.toggled.connect(self.radStandAloneToggled)
        self.radServer.toggled.connect(self.radServerToggled)
        self.radClient.toggled.connect(self.radClientToggled)

        if self.mw.settings_profile != "gui":
            self.radServer.setEnabled(False)

        tag = "info"
        if self.proxyType == ProxyType.SERVER:
            try:
                filename = os.path.join(self.getProxyServerDir(), "mediamtx.yml") 
                with open(filename, 'r') as file:
                    content = file.read()
                    key = "logLevel:"
                    index = content.index(key)
                    eol = content.index("\n", index)
                    tag = content[index + len(key):eol].strip()
            except Exception as ex:
                logger.error(f"Media MTX configuration file read error: {ex}")
        self.cmbProxyLog.setCurrentText(tag)
        self.cmbProxyLog.currentTextChanged.connect(self.cmbProxyLogChanged)

        lytProxyType = QGridLayout(self.grpProxyType)
        lytProxyType.addWidget(self.radStandAlone,       0, 0, 1, 1)
        lytProxyType.addWidget(QLabel(),                 1, 0, 1, 2)
        lytProxyType.addWidget(self.radClient,           2, 0, 1, 1)
        lytProxyType.addWidget(self.lblRemote,           2, 1, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytProxyType.addWidget(self.btnUpdate,           3, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytProxyType.addWidget(self.txtRemote,           3, 1, 1, 1)
        lytProxyType.addWidget(self.chkListen,           4, 0, 1, 2)
        lytProxyType.addWidget(pnlTimeout,          5, 0, 1, 2)
        #lytProxyType.addWidget(QLabel(),                 6, 0, 1, 1)
        lytProxyType.addWidget(self.radServer,           7, 0, 1, 1)
        lytProxyType.addWidget(self.lblConnect,          7, 1, 1, 1)
        lytProxyType.addWidget(self.lblServer,           8, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)

        lytProxyType.addWidget(self.chkEnableHttpServer,  11, 0, 1, 1)

        lytProxyType.addWidget(self.grpMediaMTX,         12, 0, 1, 2)

        lytProxyType.addWidget(self.grpAlarmBroadcast,  13, 0, 1, 2)
        lytProxyType.addWidget(QLabel(),                15, 0, 1, 1)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.grpProxyType,  0, 0, 1, 2)

        lytMain.addWidget(QLabel(),           2, 0, 1, 2)
        lytMain.setRowStretch(2, 10)

    def getProxyServerDir(self):
        dir = None
        autoDownload = self.chkAutoDownload.isChecked()
        if autoDownload:
            dir = os.path.join(self.mw.getCacheLocation(), "proxy")
        else:
            dir = self.txtDirectoryMTX.text()
        return dir

    def setProxyType(self, type):
        if type == self.proxyType:
            return
        if self.mw.pm.countPlayers():
            ret = QMessageBox.question(self.mw, "Closing Streams", "Continuing with this operation will close all open streams and may require the application to restart.\n\nAre you sure you want to do this?")
            if ret == QMessageBox.StandardButton.No:
                match (self.proxyType):
                    case ProxyType.CLIENT:
                        self.radClient.setChecked(True)
                    case ProxyType.SERVER:
                        self.radServer.setChecked(True)
                    case ProxyType.STAND_ALONE:
                        self.radStandAlone.setChecked(True)
                return
            self.mw.closeAllStreams()

        self.proxyType = type
        self.mw.settings.setValue(self.proxyTypeKey, type)
        self.mw.stopProxyServer()
        self.mw.stopOnvifServer()
        self.mw.stopHttpServer()

        self.grpAlarmBroadcast.setEnabled(type == ProxyType.SERVER)
        self.grpMediaMTX.setEnabled(type == ProxyType.SERVER)
        self.chkEnableHttpServer.setEnabled(type == ProxyType.SERVER)

        if type == ProxyType.SERVER and self.mw.settings_profile == "gui":
            self.setServersLabel()
            self.mw.manageBroadcaster(self.getInterfaces())
            self.mw.startProxyServer(self.chkAutoDownload.isChecked(), self.getProxyServerDir())
            if self.chkEnableHttpServer.isChecked():
                self.mw.startHttpServer()
            self.mw.startOnvifServer("")

        if not hasattr(self.mw, "cameraPanel"):
            return

        if lstCamera := self.mw.cameraPanel.lstCamera:
            cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
            for camera in cameras:
                self.mw.addCameraProxy(camera)
                if type == ProxyType.STAND_ALONE:
                    camera.onvif_data.nullifyGetProxyURI()
                else:
                    camera.onvif_data.getProxyURI = self.mw.getProxyURI
                for profile in camera.profiles:
                    if type == ProxyType.STAND_ALONE:
                        profile.nullifyGetProxyURI()
                    else:
                        profile.getProxyURI = self.mw.getProxyURI

        if type == ProxyType.CLIENT:
            self.manageAnalyzers(self.chkListen.isChecked())
            if len(self.mw.alarm_ordinals) != self.mw.cameraPanel.lstCamera.count():
                if not self.mw.cameraPanel.allCamerasFilled():
                    QMessageBox.warning(self.mw, "Application Restart Required", "It is necessary to re-start Onvif GUI in order to enable this change", QMessageBox.StandardButton.Close)
                    self.mw.close()
                    return
                self.mw.closeAllStreams()
                self.mw.cameraPanel.lstCamera.clear()
                self.mw.cameraPanel.btnDiscoverClicked()
        else:
            self.manageAnalyzers(False)

    def radClientToggled(self, checked):
        self.lblRemote.setEnabled(checked)
        self.txtRemote.setEnabled(checked)
        self.chkListen.setEnabled(checked)
        if checked:
           self.setProxyType(ProxyType.CLIENT)

    def radStandAloneToggled(self, checked):
        if checked:
            self.setProxyType(ProxyType.STAND_ALONE)

    def radServerToggled(self, checked):
        self.lblConnect.setEnabled(checked)
        self.chkAutoDownload.setEnabled(checked)
        if checked:
            self.setProxyType(ProxyType.SERVER)
            try:
                self.mw.cameraPanel.btnDiscoverClicked()
            except Exception as ex:
                pass
        else:
            self.lblServer.setText("\n\n")

    def setServersLabel(self):
        lbl = ""
        for if_addr in self.if_addrs:
            lbl += f'rtsp://{if_addr}:8554/\n'
        self.lblServer.setText(lbl)

    def txtRemoteEdited(self, arg):
        self.mw.settings.setValue(self.proxyRemoteKey, arg)
        comps = arg[:len(arg)-1][7:].split(":")
        ip_addr = comps[0]

        try:
            ip = ipaddress.ip_address(ip_addr)
            if self.txtRemote.text() != self.proxyRemote:
                self.btnUpdate.setEnabled(True)
        except ValueError:
            self.btnUpdate.setEnabled(False)

    def btnUpdateClicked(self):
        try:
            if self.mw.pm.countPlayers():
                QMessageBox.information(self.mw, "Closing Streams", "All current streams will be closed")
                self.mw.closeAllStreams()

            arg = self.txtRemote.text()
            comps = arg[:len(arg)-1][7:].split(":")
            ip_addr = comps[0]
            port = 8550
            if not self.mw.client:
                self.mw.initializeClient(ip_addr)
            else:
                self.mw.client.setEndpoint(ip_addr, port)
            self.proxyRemote = self.txtRemote.text()

            if lstCamera := self.mw.cameraPanel.lstCamera:
                cameras = [lstCamera.item(x) for x in range(lstCamera.count())]
                for camera in cameras:
                    self.mw.addCameraProxy(camera)

            self.mw.startListener(self.getInterfaces())

            self.btnUpdate.setEnabled(False)
        except Exception as ex:
            QMessageBox.critical(self.mw, "Proxy Client Error", str(ex))

    def chkAutoDownloadChecked(self, state):
        self.mw.settings.setValue(self.autoDownloadKey, state)
        self.txtDirectoryMTX.setEnabled(not self.chkAutoDownload.isChecked())

    def chkListenChecked(self, state):
        self.mw.settings.setValue(self.listenKey, state)
        self.manageAnalyzers(state)

    def chkEnableHttpServerChecked(self, state):
        self.mw.settings.setValue(self.httpKey, state)
        if state:
            self.createMtxIndex()
            self.mw.startHttpServer()
        else:
            self.mw.stopHttpServer()

    def spnTimeoutChanged(self, value):
        print("spnTimeoutChanged", value)
        self.mw.settings.setValue(self.timeoutKey, value)
        self.mw.client.timeout = value

    def grpAlarmBroadcastClicked(self):
        self.mw.settings.setValue(self.broadcastEnableKey, int(self.grpAlarmBroadcast.isChecked()))
        self.mw.manageBroadcaster(self.getInterfaces())

    def cmbInterfacesChanged(self, interface):
        self.mw.settings.setValue(self.broadcastInterfaceKey, interface)
        self.mw.manageBroadcaster(self.getInterfaces())

    def generateAlarmsLocally(self):
        if self.proxyType == ProxyType.CLIENT and self.chkListen.isChecked():
            return False
        else:
            return True
        
    def getTabIndexByName(self, name):
        index = -1
        for i in range(self.mw.tab.count()):
            if self.mw.tab.tabText(i) == name:
                return i
        return index
        
    def manageAnalyzers(self, state):
        if state:
            self.mw.tab.removeTab(self.getTabIndexByName("Video"))
            self.mw.tab.removeTab(self.getTabIndexByName("Audio"))
            self.mw.videoConfigure = None
            self.mw.audioConfigure = None
            self.mw.startListener(self.getInterfaces())
        else:
            self.mw.stopListener()
            self.mw.tab.addTab(self.mw.videoPanel, "Video")
            self.mw.tab.addTab(self.mw.audioPanel, "Audio")
            if not self.mw.videoConfigure:
                videoWorkerName = self.mw.videoPanel.cmbWorker.currentText()
                if len(videoWorkerName):
                    self.mw.loadVideoConfigure(videoWorkerName)
            if not self.mw.audioConfigure:
                audioWorkerName = self.mw.audioPanel.cmbWorker.currentText()
                if len(audioWorkerName):
                    self.mw.loadAudioConfigure(audioWorkerName)

    def getInterfaces(self):
        if_addrs = []
        match self.proxyType:
            case ProxyType.SERVER:
                if self.grpAlarmBroadcast.isChecked():
                    if_addrs.append(self.cmbInterfaces.currentText())
            case ProxyType.CLIENT:
                if_addrs = self.if_addrs
        return if_addrs
    
    def cmbProxyLogChanged(self, arg):
        try:
            filename = os.path.join(self.getProxyServerDir(), "mediamtx.yml") 
            with open(filename, 'r') as file:
                content = file.read()

            key = "logLevel:"
            index = content.index(key)
            eol = content.index("\n", index)
            tag = content[index:eol].strip()
            content = content.replace(tag, f'logLevel: {arg}')

            with open(filename, 'w') as file:
                file.write(content) 

        except Exception as ex:
            logger.error(f"Media MTX configuration file read error: {ex}")
        
    def setMediaMTXProxies(self):
        try:
            dir = self.getProxyServerDir()

            config_filename = f'{dir}/mediamtx.yml'
            if not os.path.isfile(config_filename):
                self.mw.signals.error.emit(f'Invalid path for MediaMTX configuration file: {config_filename}')
                return

            uri_map = {}
            for proxy_key in self.mw.proxies:
                username = None
                password = None
                if camera := self.mw.cameraPanel.getCamera(self.mw.proxies[proxy_key]):
                    if profile := camera.getDisplayProfile():
                        username = profile.username()
                        password = profile.password()
                mtx_key = self.mw.proxies[proxy_key][len("rtsp://"):]
                mtx_key = mtx_key[mtx_key.find('/')+1:]

                uri = None
                if username and password:
                    uri = f'rtsp://{username}:{password}@{proxy_key[len("rtsp://"):]}'
                else:
                    uri = proxy_key

                uri_map[mtx_key] = uri                

            mtx_uris = {}
            relevant = False
            key = None
            value = None
            with open(config_filename, 'r') as config:
                for readline in config:
                    line = readline.strip()

                    if relevant:
                        if not line.startswith("#"):
                            if not key:
                                key = line[:-1]
                            else:
                                value = line[len("source: "):]
                                mtx_uris[key] = value
                                key = None
                                value = None

                    if line.startswith("paths:"):
                        relevant = True

                    if line.startswith("all_others"):
                        relevant = False

            if mtx_uris == uri_map:
                return

            relevant = False
            lines = []
            with open(config_filename, 'r') as config:
                for readline in config:
                    line = readline.strip()

                    if not relevant or line.startswith("#"):
                        lines.append(readline)

                    if line.startswith("paths:"):
                        relevant = True

                        for key in uri_map:
                            lines.append(f'  {key}:\n')
                            lines.append(f'    source: {uri_map[key]}\n')

                    if line.startswith("all_others"):
                        lines.append(readline)
                        relevant = False

            
            with open(config_filename, 'w') as config:
                for line in lines:
                    config.write(line)

            self.createMtxIndex()

        except Exception as ex:
            logger.error(f'Error setting Media MTX proxies: {ex}')
            self.mw.signals.error.emit(f'Error setting Media MTX proxies: {ex}')

    def createMtxIndex(self):
        try:
            links = {}
            lst = self.mw.cameraPanel.lstCamera
            cameras = [lst.item(x) for x in range(lst.count())]
            for camera in cameras:
                #print("camera", camera.name())
                for profile in camera.profiles:
                    #print("    profile:", f"{camera.serial_number()}/{profile.profile()}")
                    links[f"{camera.name()} ({profile.profile()})"] = f"{camera.serial_number()}/{profile.profile()}"

            #for link in links:
            #    print("Link", link, links[link])

            #print(self.generateIndexHtml(links))

            with open(f'{self.getProxyServerDir()}/index.html', 'w') as file:
                file.write(self.generateIndexHtml(links))

        except Exception as ex:
            logger.error(f"error generating MTX index.html: {ex}")

    def generateIndexHtml(self, cameras, webrtc_port=8889):
        links_html = "\n".join(
            f'    <li><a class="cam" data-path="{html.escape(path)}">'
            f'{html.escape(name)}</a></li>'
            for name, path in cameras.items()
        )

        return f"""<!DOCTYPE html>
    <html>
    <head>
    <meta charset="utf-8">
    <title>Cameras</title>
    </head>
    <body>

    <h1>Camera List</h1>

    <ul>
    {links_html}
    </ul>

    <script>
    const webrtcPort = {webrtc_port};
    const loc = window.location;

    document.querySelectorAll(".cam").forEach(a => {{
        const path = a.dataset.path;
        a.href = `${{loc.protocol}}//${{loc.hostname}}:${{webrtcPort}}/${{path}}`;
    }});
    </script>

    </body>
    </html>
    """



from PyQt6.QtWidgets import QGridLayout, QLabel, \
    QDialog, QPushButton
from PyQt6.QtCore import Qt

class InfoDialog(QDialog):
    def __init__(self, p):
        super().__init__(p)
        self.setWindowTitle("info")
        self.lblMessage = QLabel()
        self.btnOK = QPushButton("OK")
        self.btnOK.clicked.connect(self.hide)
        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lblMessage, 0, 0, 1, 1)
        lytMain.addWidget(QLabel(),        1, 0, 1, 1)
        lytMain.addWidget(self.btnOK,      2, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)



import os
import sys
from pathlib import Path
import importlib.util

directory = Path(__file__).parent.absolute()
name = os.path.split(directory)[-1] 

module_ext = ".so"
if sys.platform == "win32":
    module_ext = ".pyd"
    os.add_dll_directory(directory)

for file in os.listdir(directory):
    filename, ext = os.path.splitext(file)
    if ext == module_ext and filename.startswith(name):
        target = os.path.join(directory, file)
        spec = importlib.util.spec_from_file_location(name, target)
        sys.modules[name] = importlib.util.module_from_spec(spec)



#*******************************************************************************
# onvif-gui/setup.py
#
# Copyright (c) 2023, 2024 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

from setuptools import setup, find_packages

with open("README.md", "r", encoding = 'cp850') as fh:
    long_description = fh.read()

setup(
    name="onvif-gui",
    version="3.3.7",
    author="Stephen Rhodes",
    author_email="sr99622@gmail.com",
    description="GUI program for onvif",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires='>=3.10',
    entry_points={
        'gui_scripts': [
            'onvif-gui=onvif_gui.main:run'
        ]
    }
)


#/********************************************************************
# onvif-gui/panels/video/modules/yolox.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

IMPORT_ERROR = ""
MODULE_NAME = "yolox"

try:
    import os
    import sys
    from loguru import logger
    import numpy as np
    from pathlib import Path
    from onvif_gui.components import ComboSelector, FileSelector, ThresholdSlider, TargetSelector
    from PyQt6.QtWidgets import QWidget, QGridLayout, QLabel, QCheckBox, QMessageBox, \
        QGroupBox, QSpinBox
    from onvif_gui.panels.video.modules.settings import VideoModelSettings
    import time
    import torch
    import requests
    from torchvision.transforms import functional
    import torch.nn as nn
    from onvif_gui.panels.video.modules.yolox.models import YOLOX, YOLOPAFPN, YOLOXHead
    import onvif_gui.panels.video.modules.yolox.utils
    if sys.platform != "darwin":
        import openvino as ov

except ModuleNotFoundError as ex:
    IMPORT_ERROR = str(ex)
    logger.debug("Import Error has occurred, missing modules need to be installed, please consult documentation: ", ex)

os.environ['KMP_DUPLICATE_LIB_OK']='True'

class VideoConfigure(QWidget):
    def __init__(self, mw):
        try:
            super().__init__()
            self.mw = mw
            self.name = MODULE_NAME
            self.camera = None
            self.initialized = False
            self.autoKey = "Module/" + MODULE_NAME + "/autoDownload"

            if len(IMPORT_ERROR):
                self.mw.videoPanel.lblCamera.setText(f'Configuration error - {IMPORT_ERROR}')
                return

            self.chkAuto = QCheckBox("Automatically download model")
            self.chkAuto.setChecked(int(self.mw.settings.value(self.autoKey, 1)))
            self.chkAuto.stateChanged.connect(self.chkAutoClicked)
            self.txtFilename = FileSelector(mw, MODULE_NAME)
            self.txtFilename.setEnabled(not self.chkAuto.isChecked())
            self.cmbRes = ComboSelector(mw, "Size", ("160", "320", "480", "640", "960", "1280"), "640", MODULE_NAME)
            self.cmbModelName = ComboSelector(mw, "Name", ("yolox_tiny", "yolox_s", "yolox_m", "yolox_l", "yolox_x"), "yolox_s", MODULE_NAME)
            
            apis = None
            if sys.platform == "darwin":
                apis = ["PyTorch"]
                default_api = "PyTorch"
            elif sys.platform == "win32":
                apis = ["OpenVINO", "PyTorch"]
                default_api = "OpenVINO"
            else:
                apis = ["PyTorch", "OpenVINO"]            
                if self.mw.haveNvidia():
                    default_api = "PyTorch"
                else:
                    default_api = "OpenVINO"
            
            self.cmbAPI = ComboSelector(mw, "API", apis, default_api, MODULE_NAME)
            self.cmbAPI.cmbBox.currentTextChanged.connect(self.cmbAPIChanged)

            self.cmbDevice = ComboSelector(mw, "Device", self.getDevices(self.cmbAPI.currentText()), "AUTO", MODULE_NAME)

            self.sldConfThre = ThresholdSlider(mw, "Confidence", MODULE_NAME)
            self.selTargets = TargetSelector(self.mw, MODULE_NAME)

            self.spnSkipFrames = QSpinBox()
            self.lblSkipFrames = QLabel("Skip Frames")
            self.spnSkipFrames.setValue(0)
            self.spnSkipFrames.valueChanged.connect(self.spnSkipFramesChanged)

            self.spnSampleSize = QSpinBox()
            self.lblSampleSize = QLabel("Sample Size")
            self.spnSampleSize.setMinimum(1)
            self.spnSampleSize.setValue(1)
            self.spnSampleSize.valueChanged.connect(self.spnSampleSizeChanged)

            grpSystem = QGroupBox("System wide model parameters")
            lytSystem = QGridLayout(grpSystem)
            lytSystem.addWidget(self.chkAuto,      0, 0, 1, 4)
            lytSystem.addWidget(self.txtFilename,  1, 0, 1, 4)
            lytSystem.addWidget(self.cmbModelName, 2, 0, 1, 2)
            lytSystem.addWidget(self.cmbRes,       2, 2, 1, 2)
            lytSystem.addWidget(self.cmbAPI,       3, 0, 1, 2)
            lytSystem.addWidget(self.cmbDevice,    3, 2, 1, 2)

            self.grpCamera = QGroupBox("Check to enable")
            self.grpCamera.setCheckable(True)
            self.grpCamera.clicked.connect(self.mw.cameraPanel.tabVideo.chkAnalyzeVideoChecked)
            lytCamera = QGridLayout(self.grpCamera)
            lytCamera.addWidget(self.sldConfThre,    0, 0, 1, 4)
            lytCamera.addWidget(self.lblSkipFrames,  1, 0, 1, 1)
            lytCamera.addWidget(self.spnSkipFrames,  1, 1, 1, 1)
            lytCamera.addWidget(self.lblSampleSize,  1, 2, 1, 1)
            lytCamera.addWidget(self.spnSampleSize,  1, 3, 1, 1)
            lytCamera.addWidget(QLabel(),            2, 0, 1, 4)
            lytCamera.addWidget(self.selTargets,     3, 0, 1, 4)

            lytMain = QGridLayout(self)
            lytMain.addWidget(grpSystem,         0, 0, 1, 1)
            lytMain.addWidget(QLabel(),          1, 0, 1, 1)
            lytMain.addWidget(self.grpCamera,    2, 0, 1, 1)
            lytMain.addWidget(QLabel(),          3, 0, 1, 1)
            lytMain.setRowStretch(3, 10)

            self.enableControls(False)
            if camera := self.mw.cameraPanel.getCurrentCamera():
                self.setCamera(camera)

            self.initialized = True

        except Exception as ex:
            logger.exception(MODULE_NAME + " configure failed to load")
            QMessageBox.critical(None, f'{MODULE_NAME} Error', f'{MODULE_NAME} configure failed to initialize: {ex}')

    def chkAutoClicked(self, state):
        self.mw.settings.setValue(self.autoKey, state)
        self.txtFilename.setEnabled(not self.chkAuto.isChecked())

    def spnSkipFramesChanged(self, value):
        if self.camera:
            if self.camera.videoModelSettings:
                self.camera.videoModelSettings.setSkipFrames(value)

    def spnSampleSizeChanged(self, value):
        if self.camera:
            if self.camera.videoModelSettings:
                self.camera.videoModelSettings.setSampleSize(value)

        self.selTargets.sldGain.setMaximum(value)

    def getDevices(self, api):
        devices = []
        if api == "OpenVINO" and sys.platform != "darwin":
            devices = ["AUTO"] + ov.Core().available_devices
        if api == "PyTorch":
            devices = ["auto", "cpu"]
            if torch.cuda.is_available():
                devices.append("cuda")
            if torch.backends.mps.is_available():
                devices.append("mps")
        return devices

    def cmbAPIChanged(self, text):
        self.cmbDevice.clear()
        self.cmbDevice.addItems(self.getDevices(text))

    def setCamera(self, camera):
        if not camera:
            return
        
        self.camera = camera
        if camera.videoModelSettings.module_name != MODULE_NAME:
            camera.videoModelSettings = VideoModelSettings(self.mw, camera, MODULE_NAME)

        if not len(IMPORT_ERROR):
            self.mw.videoPanel.lblCamera.setText(f'Camera - {camera.name()}')
            self.sldConfThre.setValue(camera.videoModelSettings.confidence)
            self.spnSkipFrames.setValue(camera.videoModelSettings.skipFrames)
            self.spnSampleSize.setValue(camera.videoModelSettings.sampleSize)
            self.selTargets.setModelParameters(camera.videoModelSettings)
            if profile := self.mw.cameraPanel.getProfile(camera.uri()):
                self.enableControls(profile.getAnalyzeVideo())

    def enableControls(self, state):
        self.grpCamera.setChecked(bool(state))
        self.grpCamera.setEnabled(False)

        if camera := self.mw.cameraPanel.getCurrentCamera():
            self.grpCamera.setEnabled(True)
            if state:
                self.grpCamera.setTitle("Video analysis enabled")
            else:
                self.grpCamera.setTitle("Check to enable")

class VideoWorker:
    def __init__(self, mw):
        try:
            self.mw = mw
            self.last_ex = ""
            self.api = self.mw.videoConfigure.cmbAPI.currentText()

            if self.mw.videoConfigure.name != MODULE_NAME or len(IMPORT_ERROR) > 0 or self.mw.glWidget.model_loading:
                return
            
            self.mw.glWidget.model_loading = True
            time.sleep(1)

            self.mw.videoConfigure.selTargets.indAlarm.setState(0)
            self.mw.videoConfigure.selTargets.barLevel.setLevel(0)

            self.torch_device = None
            self.torch_device_name = None
            self.ov_device = None
            self.compiled_model = None
            ov_model = None

            self.num_classes = 80
            self.res = int(self.mw.videoConfigure.cmbRes.currentText())
            initializer_data = torch.rand(1, 3, self.res, self.res)
            self.model_name = self.mw.videoConfigure.cmbModelName.currentText()

            if self.api == "PyTorch":
                self.torch_device_name = self.mw.videoConfigure.cmbDevice.currentText()
            if self.api == "OpenVINO":
                self.ov_device = self.mw.videoConfigure.cmbDevice.currentText()

            if self.api == "OpenVINO" and Path(self.get_ov_model_filename()).is_file() and sys.platform != "darwin":
                ov_model = ov.Core().read_model(self.get_ov_model_filename())
            
            if (self.api == "OpenVINO" and not ov_model) or self.api == "PyTorch":
                self.torch_device_name = "cpu"
                if self.api == "PyTorch":
                    if torch.cuda.is_available():
                        self.torch_device_name = "cuda"
                    if torch.backends.mps.is_available():
                        self.torch_device_name = "mps"
                    if self.mw.videoConfigure.cmbDevice.currentText() == "cpu":
                        self.torch_device_name = "cpu"
                self.torch_device = torch.device(self.torch_device_name)

                size = {'yolox_tiny': [0.33, 0.375],
                        'yolox_s': [0.33, 0.50], 
                        'yolox_m': [0.67, 0.75],
                        'yolox_l': [1.00, 1.00],
                        'yolox_x': [1.33, 1.25]}[self.model_name]

                self.model = None
                self.model = self.get_model(self.num_classes, size[0], size[1], None).to(self.torch_device)
                self.model.eval()

                self.ckpt_file = None
                if self.mw.videoConfigure.chkAuto.isChecked():
                    self.ckpt_file = self.get_auto_ckpt_filename()
                    cache = Path(self.ckpt_file)
                    if not cache.is_file():
                        self.mw.signals.showWaitDialog.emit("Please wait for yolox model to download")
                        try:
                            cache.parent.mkdir(parents=True, exist_ok=True)
                            id = f'{self.model_name}.pth'
                            link = "https://github.com/Megvii-BaseDetection/YOLOX/releases/download/0.1.1rc0/" + id
                            response = requests.get(link, allow_redirects=True, timeout=(10, 120))
                            if not response:
                                raise RuntimeError(f'Error downloading {link}: {response.status_code}')
                            with open(self.ckpt_file, 'wb') as content:
                                content.write(response.content)
                            if os.path.isfile(self.ckpt_file):

                                hashes = {
                                    "yolox_tiny.pth": "9de513de589ac98bb92d3bca53b5af7b9acfa9b0bacb831f7999d0f7afaee8f0",
                                    "yolox_s.pth": "f55ded7181e1b0c13285c56e7790b8f0e8f8db590fe4edb37f0b7f345c913a30",
                                    "yolox_m.pth": "60076992b32da82951c90cfa7bd6ab70eba9eda243e08b940a396f60ac2d19b6",
                                    "yolox_l.pth": "1e6b7fa6240375370b2a8a8eab9066b3cdd43fd1d0bfa8d2027fd3a51def2917",
                                    "yolox_x.pth": "5652330b6ae860043f091b8f550a60c10e1129f416edfdb65c259be6caf355cf"
                                }

                                verified = False
                                if hash := self.mw.calculate_sha256(self.ckpt_file):
                                    if hash == hashes.get(id, None):
                                        verified = True
                                if not verified:
                                    os.remove(self.ckpt_file)
                                    raise RuntimeError(f'Error verifying {self.ckpt_file}')

                                logger.debug(f'model {self.ckpt_file} was downloaded succesfully')
                        except Exception as ex:
                            logger.error(f'YOLO model {link} download failure: {ex}')
                        self.mw.signals.hideWaitDialog.emit()
                else:
                    self.ckpt_file = self.mw.videoConfigure.txtFilename.text()

                self.model.load_state_dict(torch.load(self.ckpt_file, map_location="cpu")["model"])
                self.model(initializer_data.to(self.torch_device))

            if self.api == "OpenVINO" and sys.platform != "darwin":
                if not ov_model:
                    ov_model = ov.convert_model(self.model, example_input=initializer_data)
                    ov.save_model(ov_model, self.get_ov_model_filename())

                self.ov_device = self.mw.videoConfigure.cmbDevice.currentText()
                if self.ov_device != "CPU":
                    ov_model.reshape({0: [1, 3, self.res, self.res]})
                ov_config = {}
                if "GPU" in self.ov_device or ("AUTO" in self.ov_device and "GPU" in ov.Core().available_devices):
                    ov_config = {"GPU_DISABLE_WINOGRAD_CONVOLUTION": "YES"}

                self.compiled_model = ov.compile_model(ov_model, self.ov_device, ov_config)
                self.compiled_model(initializer_data)

            if not self.torch_device:
                self.torch_device = torch.device("cpu")

        except Exception as ex:
            logger.exception(MODULE_NAME + " initialization failure")
            self.mw.signals.error.emit(f'{MODULE_NAME} initialization failure - {ex}')

        self.mw.glWidget.model_loading = False

    def preprocess(self, frame):
        h = frame.shape[0]
        w = frame.shape[1]

        test_size = (self.res, self.res)
        ratio = min(test_size[0] / h, test_size[1] / w)
        inf_shape = (int(h * ratio), int(w * ratio))
        bottom = test_size[0] - inf_shape[0]
        side = test_size[1] - inf_shape[1]
        pad = (0, 0, side, bottom)

        timg = functional.to_tensor(frame).to(self.torch_device)
        timg *= 255
        timg = functional.resize(timg, inf_shape)
        timg = functional.pad(timg, pad, 114)
        timg = timg.unsqueeze(0)
        return timg
    
    def postprocess(self, outputs, confthre, size, targets):
        nmsthre = 0.65
        if isinstance(outputs, np.ndarray):
            outputs = torch.from_numpy(outputs)
        outputs = onvif_gui.panels.video.modules.yolox.utils.postprocess(outputs, self.num_classes, confthre, nmsthre)
        output = None
        boxes = []
        test_size = (self.res, self.res)

        h = size[0]
        w = size[1]
        ratio = min(test_size[0] / h, test_size[1] / w)

        if outputs[0] is not None:
            output = outputs[0].cpu().numpy().astype(float)
            output[:, 0:4] /= ratio
            output[:, 4] *= output[:, 5]
            output = np.delete(output, 5, 1)

            labels = output[:, 5].astype(int)
            for i in range(len(labels)):
                if labels[i] in targets:
                    boxes.append(output[i, 0:4])

        return boxes

    def __call__(self, frame, uri):
        try:
            if len(IMPORT_ERROR) or self.mw.glWidget.model_loading or not self.mw.videoConfigure:
                return
            
            if frame is None:
                return
            
            if self.mw.videoConfigure.name != MODULE_NAME:
                self.mw.videoConfigure.selTargets.barLevel.setLevel(0)
                self.mw.videoConfigure.selTargets.indAlarm.setState(0)
                return
            
            camera = self.mw.cameraPanel.getCamera(uri)
            if not camera:
                return
            if camera.videoModelSettings.module_name != MODULE_NAME:
                camera.videoModelSettings = VideoModelSettings(self.mw, camera, MODULE_NAME)
            
            if self.parameters_changed():
                self.__init__(self.mw)
                return
            
            h = frame.shape[0]
            w = frame.shape[1]

            if self.api == "OpenVINO" and self.compiled_model:
                timg = self.preprocess(frame)                
                infer_request = self.compiled_model.create_infer_request()
                infer_request.set_input_tensor(ov.Tensor(np.asarray(timg)))
                infer_request.infer()
                outputs = infer_request.get_output_tensor(0).data
            if self.api == "PyTorch" and self.model:
                timg = self.preprocess(frame)
                with torch.no_grad():
                    outputs = self.model(timg)

            return self.postprocess(outputs, camera.videoModelSettings.confidence / 100, [h, w], camera.videoModelSettings.targets)

        except Exception as ex:
            if self.last_ex != str(ex) and self.mw.videoConfigure.name == MODULE_NAME:
                logger.exception(f'{MODULE_NAME} runtime error - {ex}')
            self.last_ex = str(ex)

    def parameters_changed(self):
        result = False

        api = self.api == self.mw.videoConfigure.cmbAPI.currentText()
        name = self.model_name == self.mw.videoConfigure.cmbModelName.currentText()
        res = str(self.res) == self.mw.videoConfigure.cmbRes.currentText()

        dev = False
        if self.api == "PyTorch":
            dev = True
            if self.mw.videoConfigure.cmbDevice.currentText() != "auto":
                dev = self.torch_device_name == self.mw.videoConfigure.cmbDevice.currentText()
        if self.api == "OpenVINO":
            dev = self.ov_device == self.mw.videoConfigure.cmbDevice.currentText()

        if not api or not name or not res or not dev:
            result = True

        return result

    def get_ov_model_filename(self):
        model_name = self.mw.videoConfigure.cmbModelName.currentText()
        openvino_device = self.mw.videoConfigure.cmbDevice.currentText()
        path = os.path.join(self.mw.getCacheLocation(), "checkpoints", model_name, openvino_device, "model.xml")
        return path

    def get_auto_ckpt_filename(self):
        model_name = self.mw.videoConfigure.cmbModelName.currentText()
        path = os.path.join(self.mw.getCacheLocation(), "checkpoints", f'{model_name}.pth')
        return path

    def get_model(self, num_classes, depth, width, act):
        def init_yolo(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03

        in_channels = [256, 512, 1024]
        backbone = YOLOPAFPN(depth, width, in_channels=in_channels)
        head = YOLOXHead(num_classes, width, in_channels=in_channels)
        model = YOLOX(backbone, head)

        model.apply(init_yolo)
        model.head.initialize_biases(1e-2)
        return model



#*******************************************************************************
# onvif-gui/run.py
#
# Copyright (c) 2023 Stephen Rhodes 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#******************************************************************************/

import sys
from PyQt6.QtWidgets import QApplication
from onvif_gui import MainWindow

clear_settings = False
if len(sys.argv) > 1:
    if str(sys.argv[1]) == "--clear":
        clear_settings = True

app = QApplication(sys.argv)

app.setStyle('Fusion')
window = MainWindow(clear_settings)

window.show()
app.exec()



#/********************************************************************
# onvif-gui/onvif_gui/panels/cameras/datastructures.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QListWidgetItem
from PyQt6.QtCore import Qt, pyqtSignal, QObject, QTimer
from PyQt6.QtGui import QIcon, QColor
import libonvif as onvif
from onvif_gui.panels.camera.systemtab import SystemTabSettings
#from onvif_gui.panels.video.modules.common.yolosettings import YoloSettings
from onvif_gui.panels.video.modules.settings import VideoModelSettings
from onvif_gui.enums import ProxyType, StreamState, SnapshotAuth
from loguru import logger

class SessionSignals(QObject):
    finished = pyqtSignal()

class Session(onvif.Session):
    def __init__(self, cp, interface):
        super().__init__()
        self.cp = cp
        self.interface = interface
        self.signals = SessionSignals()
        self.discovered = lambda : self.finish()
        self.getCredential = lambda D : self.cp.getCredential(D)
        self.getData = lambda D : self.cp.getData(D)
        self.infoCallback = lambda msg : self.info(msg)
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.timeout)
        self.signals.finished.connect(self.timer.stop)
        self.active = False

    def start(self):
        self.active = True
        self.startDiscover()
        self.timer.start(10000)

    def finish(self):
        self.active = False
        self.signals.finished.emit()
        self.cp.discovered()

    def timeout(self):
        self.cp.discoveryTimeout()

    def info(self, msg):
        try:
            logger.debug(msg)
        except Exception as ex:
            logger.error("session info callback exception: {ex}")

class Camera(QListWidgetItem):
    def __init__(self, onvif_data, mw):
        super().__init__(onvif_data.alias)
        self.mw = mw
        self.icnIdle = QIcon("image:idle_lo.png")
        self.icnOn = QIcon("image:record.png")
        self.icnRecord = QIcon("image:recording_hi.png")
        self.defaultForeground = self.foreground()
        self.filled = False
        self.manual_fill = False
        self.last_msg = ""

        self.assignData(onvif_data)

        module_name = "yolox"
        #if self.mw.videoPanel:
        #module_name = self.mw.videoPanel.panel.module_name    

        self.videoModelSettings = VideoModelSettings(self.mw, self, module_name)
        self.audioModelSettings = None
        self.systemTabSettings = SystemTabSettings(self.mw, self)
        self.manual_recording = False
        self.ordinalKey = f'{self.serial_number()}/Ordinal'
        self.ordinal = self.getOrdinal()
        self.volumeKey = f'{self.serial_number()}/Volume'
        self.volume = self.getVolume()
        self.muteKey = f'{self.serial_number()}/Mute'
        self.mute = self.getMute()
        self.snapshotAuthKey = f'{self.serial_number()}/SnapshotAuth'
        self.snapshotAuth = self.getSnapshotAuth()

    def assignData(self, data):
        self.onvif_data = data
        data.setSetting = self.mw.settings.setValue
        data.getSetting = self.mw.settings.value
        if self.mw.settingsPanel.proxy.proxyType != ProxyType.STAND_ALONE:
            data.getProxyURI = self.mw.getProxyURI

        for profile in data.profiles:
            profile.setSetting = self.mw.settings.setValue
            profile.getSetting = self.mw.settings.value
            if self.mw.settingsPanel.proxy.proxyType != ProxyType.STAND_ALONE:
                profile.getProxyURI = self.mw.getProxyURI
        
        self.profiles = data.profiles

    def uri(self):
        return self.onvif_data.uri()
    
    def serial_number(self):
        return self.onvif_data.serial_number()
    
    def name(self):
        return self.onvif_data.alias
    
    def xaddrs(self):
        return self.onvif_data.xaddrs()

    def hasAudio(self):
        return bool(self.onvif_data.audio_bitrate())
    
    def setOrdinal(self, value):
        self.ordinal = value
        self.mw.settings.setValue(self.ordinalKey, value)

    def getOrdinal(self):
        return int(self.mw.settings.value(self.ordinalKey, -1))
    
    def getMute(self):
        return bool(int(self.mw.settings.value(self.muteKey, 0)))
    
    def setMute(self, state):
        self.mute = bool(state)
        self.mw.settings.setValue(self.muteKey, int(state))

    def getVolume(self):
        return int(self.mw.settings.value(self.volumeKey, 80))
    
    def setVolume(self, volume):
        self.volume = volume
        self.mw.settings.setValue(self.volumeKey, volume)

    def getSnapshotAuth(self):
        return self.mw.settings.value(self.snapshotAuthKey, SnapshotAuth.UNKNOWN)

    def setSnapshotAuth(self, auth):
        self.snapshotAuth = auth
        self.mw.settings.setValue(self.snapshotAuthKey, auth)

    def isRunning(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        if len(players):
            result = True
        return result
    
    def isRecording(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        for player in players:
            if player.isRecording():
                result = True
        return result
    
    def isAlarming(self):
        result = False
        players = self.mw.pm.getStreamPairPlayers(self.uri())
        for player in players:
            if player.alarm_state:
                result = True
        return result
    
    def isCurrent(self):
        result = False
        for profile in self.profiles:
            camera = self.mw.cameraPanel.getCurrentCamera()
            if camera.getProfile(profile.uri()):
                result = True
        return result

    def editing(self):
        return self.flags() & Qt.ItemFlag.ItemIsEditable
    
    def setIconIdle(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnIdle)

    def setIconOn(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnOn)

    def setIconRecord(self):
        if not self.flags() & Qt.ItemFlag.ItemIsEditable:
            self.setIcon(self.icnRecord)
        
    def dimForeground(self):
        self.setForeground(QColor("#808D9E"))

    def restoreForeground(self):
        self.setForeground(self.defaultForeground)

    def isCurrent(self):
        result = False
        current_camera = self.mw.cameraPanel.getCurrentCamera()
        if current_camera:
            if current_camera.serial_number() == self.serial_number():
                result = True
        return result
    
    def getStreamState(self, index):
        result = StreamState.INVALID
        if len(self.profiles) <= index:
            return result
        if profile := self.profiles[index]:
            if player := self.mw.pm.getPlayer(profile.uri()):
                if player.image:
                    result = StreamState.CONNECTED
                else:
                    result = StreamState.CONNECTING
            else:
                result = StreamState.IDLE

            if timer := self.mw.timers.get(profile.uri()):
                if timer.isActive():
                    result = StreamState.CONNECTING
        return result
    
    def profileName(self, uri):
        result = ""
        for profile in self.profiles:
            if profile.uri() == uri:
                result = profile.profile()
        return result
    
    def recordProfileIndex(self):
        return self.systemTabSettings.record_profile

    def displayProfileIndex(self):
        return self.onvif_data.displayProfile
    
    def setDisplayProfile(self, index):
        self.onvif_data.setProfile(index)
        self.mw.settings.setValue(f'{self.serial_number()}/DisplayProfile', index)
        for i, profile in enumerate(self.profiles):
            if i == index:
                profile.setHidden(False)
            else:
                profile.setHidden(True)

    def getDisplayProfileSetting(self):
        return int(self.mw.settings.value(f'{self.serial_number()}/DisplayProfile', self.displayProfileIndex()))

    def getProfile(self, uri):
        result = None
        for profile in self.profiles:
            if  profile.uri() == uri:
                result = profile
                break
        return result
    
    def getRecordProfile(self):
        result = None
        if len(self.profiles) > self.recordProfileIndex():
            result = self.profiles[self.recordProfileIndex()]
        return result
    
    def isRecordProfile(self, uri):
        result = False
        recordProfile = self.getRecordProfile()
        if recordProfile:
            if uri == recordProfile.uri():
                result = True
        return result
    
    def getDisplayProfile(self):
        result = None
        if len(self.profiles) > self.displayProfileIndex():
            result = self.profiles[self.displayProfileIndex()]
        return result
    
    def isDisplayProfile(self, uri):
        result = False
        if displayProfile := self.getDisplayProfile():
            if uri == displayProfile.uri():
                result = True
        return result


    def companionURI(self, uri):
        result = None
        recordProfile = self.getRecordProfile()
        displayProfile = self.getDisplayProfile()
        if recordProfile and displayProfile:
            if uri == recordProfile.uri():
                result = displayProfile.uri()
            if uri == displayProfile.uri():
                result = recordProfile.uri()
        return result

    def syncData(self, data):
        for profile in self.profiles:
            if profile.profile() == data.profile():
                profile.syncData(data)
        data.profiles = self.profiles
        if self.onvif_data.profile() == data.profile():
            self.onvif_data.syncData(data)
            if self.isCurrent():
                self.mw.cameraPanel.signals.fill.emit(data)


#/********************************************************************
# onvif-gui/onvif_gui/panels/settings/general.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
import sys
from PyQt6.QtWidgets import QMessageBox, QLineEdit, QSpinBox, \
    QGridLayout, QWidget, QCheckBox, QLabel, QComboBox, QPushButton, \
    QDialog, QTextEdit, QMessageBox, QFileDialog, QGroupBox, \
    QListWidget, QDialogButtonBox, QListWidgetItem
from PyQt6.QtCore import QFile, QRect, Qt, QSettings
from PyQt6.QtGui import QTextCursor, QTextOption
from loguru import logger
import webbrowser
import platform
from onvif_gui.player import Player
from onvif_gui.enums import Style, ProxyType, SnapshotAuth
from onvif_gui.panels.camera import Snapshot
import onvif_gui
from pathlib import Path
import subprocess
import avio
from time import sleep

class LogText(QTextEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWordWrapMode(QTextOption.WrapMode.NoWrap)

    def scrollToBottom(self):
        self.moveCursor(QTextCursor.MoveOperation.End, QTextCursor.MoveMode.MoveAnchor)
        self.ensureCursorVisible()

class LogDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.geometryKey = "LogDialog/geometry"
        if rect := self.mw.settings.value(self.geometryKey):
            if rect.width() and rect.height():
                self.setGeometry(rect)
        else:
            g = mw.geometry()
            rect = QRect(0, 0, g.width()-20, 480)
            rect.moveCenter(g.center())
            self.setGeometry(rect)

        self.editor = LogText(self)
        self.editor.setReadOnly(True)
        self.editor.setFontFamily("Courier")

        self.lblSize = QLabel("Log File Size:")
        self.btnArchive = QPushButton("View Archive")
        self.btnArchive.clicked.connect(self.btnArchiveClicked)
        self.btnClear = QPushButton("  Clear  ")
        self.btnClear.clicked.connect(self.btnClearClicked)
        self.btnClose = QPushButton("Close")
        self.btnClose.clicked.connect(self.btnCloseClicked)

        pnlButton = QWidget()
        lytButton = QGridLayout(pnlButton)
        lytButton.addWidget(self.btnArchive,   0, 1, 1, 1)
        lytButton.addWidget(self.btnClear,     0, 2, 1, 1)
        lytButton.addWidget(QLabel(),          0, 3, 1, 1)
        lytButton.addWidget(self.btnClose,     0, 4, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytButton.setContentsMargins(0, 0, 0, 0)
        lytButton.setColumnStretch(3, 10)
        
        panel = QWidget()
        lytPanel = QGridLayout(panel)
        lytPanel.addWidget(self.lblSize,    0, 0, 1, 1)
        lytPanel.addWidget(pnlButton,       0, 1, 1, 1)

        lyt = QGridLayout(self)
        lyt.addWidget(self.editor, 0, 0, 1, 1)
        lyt.addWidget(panel,       1, 0, 1, 1)
        lyt.setRowStretch(0, 10)

    def closeEvent(self, e):
        self.mw.settings.setValue(self.geometryKey, self.geometry())

    def readLogFile(self, path):
        data = ""
        if os.path.isfile(path):
            with open(path, 'r') as file:
                data = file.read()
        self.setWindowTitle(path)
        self.editor.setText(data)
        y = "{:.2f}".format(os.stat(path).st_size/1000000)
        self.lblSize.setText(f'Log File Size: {y}  MB    ')
        self.editor.scrollToBottom()

    def btnArchiveClicked(self):
        path = None
        if platform.system() == "Linux":
            path = QFileDialog.getOpenFileName(self, "Select File", self.windowTitle(), options=QFileDialog.Option.DontUseNativeDialog)[0]
        else:
            path = QFileDialog.getOpenFileName(self, "Select File", self.windowTitle())[0]
        if path:
            if len(path) > 0:
                self.readLogFile(os.path.realpath(path))

    def btnCloseClicked(self):
        self.close()

    def btnClearClicked(self):
        filename = self.windowTitle()
        log_filename = self.mw.getLogFilename()

        if filename == log_filename:
            ret = QMessageBox.question(self, "Preserve File", 
                                    "Would you like to preserve the current log file with a new name?",
                                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
            
            if ret == QMessageBox.StandardButton.Yes:
                path = None
                if platform.system() == "Linux":
                    path = QFileDialog.getSaveFileName(self, "Select A New File Name", self.windowTitle(), "Text (*.txt)", options=QFileDialog.Option.DontUseNativeDialog)[0]
                else:
                    path = QFileDialog.getSaveFileName(self, "Select A New File Name", self.windowTitle(), "Text (*.txt)")[0]
                if path:
                    if len(path) > 0:
                        try:
                            logger.remove(self.mw.logger_id)
                            os.rename(filename, path)
                        except Exception as ex:
                            QMessageBox.warning(self, "File Error", f'File rename error occurred {ex}')
                        self.mw.logger_id = logger.add(log_filename)
                        logger.debug("Created new log file")
                        self.editor.setText("")
                        self.readLogFile(log_filename)

            else:
                ret = QMessageBox.warning(self, "Deleting Current Log",
                                            "You are about to delete the current log.\n"
                                            "Are you sure you want to continue?",
                                            QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
                if ret == QMessageBox.StandardButton.Ok:
                    try:
                        logger.remove(self.mw.logger_id)
                        os.remove(filename)
                        self.editor.setText("")
                    except Exception as ex:
                        QMessageBox.warning(self, "File Error", f'File delete error occurred {ex}')

                    self.logger_id = logger.add(log_filename)
                    logger.debug("Created new log file")
                    self.readLogFile(log_filename)

        else:
            ret = QMessageBox.warning(self, "Deleting File",
                                        f'\n{filename}\n\n'
                                        "You are about to delete this file.\n"
                                        "Are you sure you want to continue?\n\n"
                                        "Clicking cancel will preserve the file and return you to the current log",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            if ret == QMessageBox.StandardButton.Ok:
                QFile.remove(filename)
            self.readLogFile(log_filename)
            QMessageBox.information(self, "Current Log Displayed", "The current log has been loaded into the display", QMessageBox.StandardButton.Ok)


class ProfileItem(QListWidgetItem):
    def __init__(self, name):
        super().__init__(name)
        self.original = name
        if name != "Focus" and name != "Reader":
            self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)

class ProfileDialog(QDialog):
    def __init__(self, mw):
        super().__init__(mw)
        self.mw = mw
        self.editingItem = None
        self.deleted = []

        self.setModal(True)
        self.setWindowTitle("Manage Profiles")
        self.list = QListWidget()

        self.list.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.list.itemSelectionChanged.connect(self.onItemSelectionChanged)

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.btnAdd = QPushButton("+")
        self.btnAdd.clicked.connect(self.btnAddClicked)
        self.btnDelete = QPushButton(" - ")
        self.btnDelete.clicked.connect(self.btnDeleteClicked)
        self.btnUp = QPushButton("^")
        self.btnUp.clicked.connect(self.btnUpClicked)
        self.btnDown = QPushButton("v")
        self.btnDown.clicked.connect(self.btnDownClicked)

        lytMain = QGridLayout(self)
        lytMain.addWidget(QLabel(),         0, 0, 1, 1)
        lytMain.addWidget(self.btnUp,      1, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.addWidget(self.btnDown,    2, 0, 1, 1, Qt.AlignmentFlag.AlignCenter)
        lytMain.addWidget(QLabel(),        3, 0, 1, 1)
        lytMain.addWidget(self.list,       0, 1, 4, 5)
        lytMain.addWidget(self.btnAdd,     4, 1, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytMain.addWidget(self.btnDelete,  4, 2, 1, 1, Qt.AlignmentFlag.AlignLeft)
        lytMain.addWidget(self.buttonBox,  4, 3, 1, 3)
        lytMain.setRowStretch(0, 10)
        lytMain.setRowStretch(3, 10)

    def reject(self):
        self.hide()

    def accept(self):
        self.clearProfileNames()
        self.mw.settingsPanel.general.cmbViewerProfile.clear()
        for i in range(self.list.count()):
            name = self.list.item(i).text() 
            key = f'SettingsProfile_{i}'
            self.mw.settings.setValue(key, name)
            self.mw.settingsPanel.general.cmbViewerProfile.addItem(name)
        for name in self.deleted:
            settings = QSettings("onvif", name.text())
            settings.clear()
        self.hide()

    def onItemDoubleClicked(self, item):
        self.editingItem = item
        self.list.editItem(item)
        
    def onItemSelectionChanged(self):
        self.syncGui()

    def btnAddClicked(self):
        row = self.list.currentRow()+1
        item = ProfileItem("New Profile")
        self.list.insertItem(row, item)
        self.list.setCurrentRow(row)
        self.list.itemDoubleClicked.emit(item)
        self.syncGui()

    def btnDeleteClicked(self):
        self.deleted.append(self.list.takeItem(self.list.currentRow()))
        self.syncGui()

    def btnUpClicked(self):
        row = self.list.currentRow()
        item = self.list.takeItem(row)
        self.list.insertItem(row-1, item)
        self.list.setCurrentRow(row-1)
        self.syncGui()

    def btnDownClicked(self):
        row = self.list.currentRow()
        item = self.list.takeItem(row)
        self.list.insertItem(row+1, item)
        self.list.setCurrentRow(row+1)
        self.syncGui()

    def syncGui(self):
        self.btnDelete.setEnabled(self.list.count() > 1)
        if self.list.count():
            self.btnUp.setEnabled(self.list.currentRow() > 0)
            self.btnDown.setEnabled(self.list.currentRow() < self.list.count()-1)
        if item := self.list.currentItem():
            if item.text() == "Focus" or item.text() == "Reader":
                self.btnDelete.setEnabled(False)

    def exec(self):
        self.list.clear()
        self.deleted.clear()
        cmb = self.mw.settingsPanel.general.cmbViewerProfile
        for i in range(cmb.count()):
            self.list.addItem(ProfileItem(cmb.itemText(i)))
        self.list.setCurrentRow(0)
        self.list.setFocus()
        self.syncGui()
        super().exec()

    def clearProfileNames(self):
        i = 0
        while self.mw.settings.contains(f'SettingsProfile_{i}'):
            self.mw.settings.remove(f'SettingsProfile_{i}')
            i += 1

class GeneralOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.process = None
        self.thread = None

        self.usernameKey = "settings/username"
        self.passwordKey = "settings/password"
        self.decoderKey = "settings/decoder"
        self.bufferSizeKey = "settings/bufferSize"
        self.lagTimeKey = "settings/lagTime"
        self.startFullScreenKey = "settings/startFullScreen"
        self.autoTimeSyncKey = "settings/autoTimeSync"
        self.alarmSoundFileKey = "settings/alarmSoundFile"
        self.alarmSoundVolumeKey = "settings/alarmSoundVolume"
        self.displayRefreshKey = "settings/displayRefresh"
        self.audioDriverIndexKey = "settings/audioDriverIndex"
        self.appearanceKey = "settings/appearance"
        self.displaySizeKey = "settings/displaySize"
        self.stressTestKey = "settings/stressTest"
        self.snapshotDlgKey = "settings/snapshotDlg"

        p = Player("", self)
        audioDrivers = p.getAudioDrivers()

        decoders = ["none"] + p.getHardwareDecoders()

        self.mediamtx_process = None

        self.dlgLog = None
        self.dlgProfiles = None

        self.txtUsername = QLineEdit()
        self.txtUsername.setText(mw.settings.value(self.usernameKey, ""))
        self.txtUsername.textChanged.connect(self.usernameChanged)
        lblUsername = QLabel("Common Username")
        
        self.txtPassword = QLineEdit()
        self.txtPassword.setText(mw.settings.value(self.passwordKey, ""))
        self.txtPassword.textChanged.connect(self.passwordChanged)
        lblPassword = QLabel("Common Password")
        
        self.cmbDecoder = QComboBox()
        self.cmbDecoder.addItems(decoders)
        self.cmbDecoder.setCurrentText(mw.settings.value(self.decoderKey, "none"))
        self.cmbDecoder.currentTextChanged.connect(self.cmbDecoderChanged)
        lblDecoders = QLabel("Hardware Decoder")

        self.cmbAudioDriver = QComboBox()
        self.cmbAudioDriver.addItems(audioDrivers)
        self.cmbAudioDriver.setCurrentIndex(int(mw.settings.value(self.audioDriverIndexKey, 0)))
        self.cmbAudioDriver.currentIndexChanged.connect(self.cmbAudioDriverChanged)
        lblAudioDrivers = QLabel("Audio Driver")

        self.cmbAppearance = QComboBox()
        self.cmbAppearance.addItems(["Dark", "Light"])
        self.cmbAppearance.setCurrentText(mw.settings.value(self.appearanceKey, "Dark"))
        self.cmbAppearance.currentTextChanged.connect(self.cmbAppearanceChanged)
        lblAppearance = QLabel("Appearance")

        self.chkStartFullScreen = QCheckBox("Start Full Screen")
        self.chkStartFullScreen.setChecked(bool(int(mw.settings.value(self.startFullScreenKey, 0))))
        self.chkStartFullScreen.stateChanged.connect(self.startFullScreenChecked)

        self.chkAutoTimeSync = QCheckBox("Auto Time Sync")
        self.chkAutoTimeSync.setChecked(bool(int(mw.settings.value(self.autoTimeSyncKey, 0))))
        self.chkAutoTimeSync.stateChanged.connect(self.autoTimeSyncChecked)

        self.chkStressTest = QCheckBox("Stress Test Cycle")
        self.chkStressTest.setChecked(False)
        self.chkStressTest.stateChanged.connect(self.stressTestChecked)

        self.chkSnapshotDlg = QCheckBox("Snapshot File Dlg")
        self.chkSnapshotDlg.setChecked(bool(int(mw.settings.value(self.snapshotDlgKey, 1))))
        self.chkSnapshotDlg.stateChanged.connect(self.snapshotDlgChecked)

        pnlChecks = QWidget()
        lytChecks = QGridLayout(pnlChecks)
        lytChecks.addWidget(self.chkStartFullScreen,  0, 0, 1, 1)
        lytChecks.addWidget(self.chkAutoTimeSync,     0, 1, 1, 1)
        #lytChecks.addWidget(self.chkStressTest,       1, 0, 1, 1)
        lytChecks.addWidget(self.chkSnapshotDlg,      1, 1, 1, 1)

        self.spnDisplayRefresh = QSpinBox()
        self.spnDisplayRefresh.setMinimum(1)
        self.spnDisplayRefresh.setMaximum(1000)
        self.spnDisplayRefresh.setMaximumWidth(80)
        refresh = 10
        if sys.platform == "win32":
            refresh = 20
        self.spnDisplayRefresh.setValue(int(self.mw.settings.value(self.displayRefreshKey, refresh)))
        self.spnDisplayRefresh.valueChanged.connect(self.spnDisplayRefreshChanged)
        lblDisplayRefresh = QLabel("Display Refresh Interval (ms)")

        self.btnCloseAll = QPushButton("Start All")
        self.btnCloseAll.clicked.connect(self.btnCloseAllClicked)

        self.btnShowLogs = QPushButton("Show Logs")
        self.btnShowLogs.clicked.connect(self.btnShowLogsClicked)

        self.btnHideDisplay = QPushButton("Hide Display")
        self.btnHideDisplay.clicked.connect(self.btnHideDisplayClicked)

        self.btnClearSettings = QPushButton("Clear")
        self.btnClearSettings.clicked.connect(self.btnClearSettingsClicked)

        self.btnHelp = QPushButton("Help")
        self.btnHelp.clicked.connect(self.btnHelpClicked)

        self.btnTest = QPushButton("Test")
        self.btnTest.clicked.connect(self.btnTestClicked)

        grpNewView = QGroupBox("Open New Window")
        self.btnNewView = QPushButton("Open")
        self.btnNewView.clicked.connect(self.btnNewViewClicked)
        self.cmbViewerProfile = QComboBox()

        if self.mw.settings_profile == "gui":
            names = self.getProfileNames()
            for name in names:
                self.cmbViewerProfile.addItem(name)
        else:
            grpNewView.setEnabled(False)
        
        self.btnManageProfiles = QPushButton("...")
        self.btnManageProfiles.clicked.connect(self.btnManageProfilesClicked)
        lytNewView = QGridLayout(grpNewView)
        lytNewView.addWidget(self.btnNewView,        0, 0, 1, 1, Qt.AlignmentFlag.AlignLeft)
        lytNewView.addWidget(QLabel("Profile"),      0, 1, 1, 1)
        lytNewView.addWidget(self.cmbViewerProfile,  0, 2, 1, 1)
        lytNewView.addWidget(self.btnManageProfiles, 0, 3, 1, 1, Qt.AlignmentFlag.AlignRight)
        lytNewView.setColumnStretch(2, 10)

        pnlBuffer = QWidget()
        lytBuffer = QGridLayout(pnlBuffer)
        lytBuffer.addWidget(lblDisplayRefresh,      4, 0, 1, 3)
        lytBuffer.addWidget(self.spnDisplayRefresh, 4, 3, 1, 1)
        lytBuffer.setContentsMargins(0, 0, 0, 0)

        pnlButtons = QWidget()
        lytButtons = QGridLayout(pnlButtons)
        lytButtons.addWidget(self.btnHideDisplay,   0, 0, 1, 1)
        lytButtons.addWidget(self.btnShowLogs,      0, 1, 1, 1)
        lytButtons.addWidget(self.btnHelp,          0, 2, 1, 1)
        #lytButtons.addWidget(self.btnCloseAll,      0, 0, 1, 1)
        #lytButtons.addWidget(self.btnClearSettings, 1, 2, 1, 1)
        #lytButtons.addWidget(self.btnTest,   1, 1, 1, 1)

        self.lblMemory = QLabel()

        lytMain = QGridLayout(self)
        lytMain.addWidget(lblUsername,         1, 0, 1, 1)
        lytMain.addWidget(self.txtUsername,    1, 1, 1, 1)
        lytMain.addWidget(lblPassword,         2, 0, 1, 1)
        lytMain.addWidget(self.txtPassword,    2, 1, 1, 1)
        lytMain.addWidget(lblDecoders,         3, 0, 1, 1)
        lytMain.addWidget(self.cmbDecoder,     3, 1, 1, 1)
        lytMain.addWidget(lblAudioDrivers,     4, 0, 1, 1)
        lytMain.addWidget(self.cmbAudioDriver, 4, 1, 1, 1)
        lytMain.addWidget(lblAppearance,       5, 0, 1, 1)
        lytMain.addWidget(self.cmbAppearance,  5, 1, 1, 1)
        lytMain.addWidget(pnlChecks,           6, 0, 1, 3)
        lytMain.addWidget(pnlBuffer,           7, 0, 1, 3)
        lytMain.addWidget(grpNewView,          8, 0, 1, 3)
        lytMain.addWidget(pnlButtons,          9, 0, 1, 3)
        lytMain.addWidget(self.lblMemory,     10, 0, 1, 3)
        lytMain.setRowStretch(10, 10)

    def usernameChanged(self, username):
        self.mw.settings.setValue(self.usernameKey, username)

    def passwordChanged(self, password):
        self.mw.settings.setValue(self.passwordKey, password)

    def cmbDecoderChanged(self, decoder):
        self.mw.settings.setValue(self.decoderKey, decoder)

    def cmbAudioDriverChanged(self, index):
        self.mw.settings.setValue(self.audioDriverIndexKey, index)
        QMessageBox.warning(self.mw, "Application Restart Required", "It is necessary to re-start Onvif GUI in order to enable this change")

    def cmbAppearanceChanged(self, text):
        self.mw.settings.setValue(self.appearanceKey, text)
        if text == "Dark":
            self.mw.setStyleSheet(self.mw.style(Style.DARK))
        if text == "Light":
            self.mw.setStyleSheet(self.mw.style(Style.LIGHT))

    def getProfileNames(self):
        result = []
        i = 0
        while self.mw.settings.contains(f'SettingsProfile_{i}'):
            result.append(self.mw.settings.value(f'SettingsProfile_{i}'))
            i += 1
        if not len(result):
            result.append("Reader")
            result.append("Focus")
        return result

    def autoDiscoverChecked(self, state):
        self.mw.settings.setValue(self.autoDiscoverKey, state)

    def startFullScreenChecked(self, state):
        self.mw.settings.setValue(self.startFullScreenKey, state)

    def autoTimeSyncChecked(self, state):
        self.mw.settings.setValue(self.autoTimeSyncKey, state)
        self.mw.cameraPanel.enableAutoTimeSync(state)
        if state:
            self.mw.cameraPanel.timeSync()

    def stressTestChecked(self, state):
        ...

    def snapshotDlgChecked(self, state):
        self.mw.settings.setValue(self.snapshotDlgKey, state)

    def spnDisplayRefreshChanged(self, i):
        self.mw.settings.setValue(self.displayRefreshKey, i)
        self.mw.glWidget.timer.setInterval(i)

    def cmbInterfacesChanged(self, network):
        self.mw.settings.setValue(self.interfaceKey, network)
     
    def btnCloseAllClicked(self):
        if self.btnCloseAll.text() == "Close All":
            self.mw.closeAllStreams()
        else:
            self.mw.startAllCameras()
            self.btnCloseAll.setText("Close All")

    def btnShowLogsClicked(self):
        filename = self.mw.getLogFilename()
        if not self.dlgLog:
            self.dlgLog = LogDialog(self.mw)
        self.dlgLog.readLogFile(filename)
        self.dlgLog.setModal(False)
        self.dlgLog.show()

    def btnHelpClicked(self):
        result = webbrowser.get().open("https://github.com/sr99622/libonvif#readme-ov-file")
        if not result:
            webbrowser.get().open("https://github.com/sr99622/libonvif")

    def btnHideDisplayClicked(self):
        controlSize = self.mw.split.sizes()[1]
        if self.mw.split.sizes()[0]:
            self.mw.hide()
            self.mw.settings.setValue(self.displaySizeKey, self.mw.split.sizes()[0])
            rect = self.mw.geometry()
            rect.setX(rect.x() + rect.width() - controlSize)
            self.mw.split.setSizes([0, controlSize])
            self.mw.setGeometry(rect)
            self.mw.show()
            self.btnHideDisplay.setText("Show Display")
        else:
            self.mw.hide()
            displaySize = int(self.mw.settings.value(self.displaySizeKey, 400))
            rect = self.mw.geometry()
            rect.setX(rect.x() - displaySize)
            self.mw.split.setSizes([displaySize, controlSize])
            self.mw.setGeometry(rect)
            self.mw.show()
            self.btnHideDisplay.setText("Hide Display")

    def openFocusWindow(self):
        if not self.mw.focus_window:
            proxy = None
            match self.mw.settingsPanel.proxy.proxyType:
                case ProxyType.CLIENT:
                    proxy = self.mw.settingsPanel.proxy.txtRemote.text()
                case ProxyType.SERVER:
                    proxy = self.mw.settingsPanel.proxy.lblServer.text().split()[0]
            focus_settings = QSettings("onvif-gui", "Focus")
            focus_settings.setValue("settings/proxyType", ProxyType.CLIENT)
            focus_settings.setValue("settings/proxyRemote", proxy)
            focus_settings.setValue("settings/autoDiscover", 1)
            #focus_settings.setValue(self.collapsedKey, 1)
            self.mw.focus_window = onvif_gui.main.MainWindow(settings_profile="Focus")
        self.mw.focus_window.show()

    def btnNewViewClicked(self):
        from onvif_gui.main import MainWindow
        if self.cmbViewerProfile.currentText() == "Focus":
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.STAND_ALONE:
                QMessageBox.warning(self.mw, "Feature Unavailable", "Focus Window is not available in Stand Alone Configuration, please use Proxy Server mode to enable this feature", QMessageBox.StandardButton.Ok)
                return
            self.mw.openFocusWindow()
        elif self.cmbViewerProfile.currentText() == "Reader":
            reader_settings = QSettings("onvif-gui", "Reader")
            reader_settings.setValue("filePanel/hideCameraPanel", 1)
            main_file = Path(__file__).parent.parent.parent / "main.py"
            subprocess.Popen(["python", str(main_file), "--profile", "reader"], env=os.environ.copy(), start_new_session=True, shell=True)
        else:
            window = MainWindow(settings_profile = self.cmbViewerProfile.currentText())
            window.parent_window = self.mw
            self.mw.external_windows.append(window)
            window.show()

    def btnManageProfilesClicked(self):
        if not self.dlgProfiles:
            self.dlgProfiles = ProfileDialog(self.mw)
        self.dlgProfiles.exec()

    def btnClearSettingsClicked(self):
        ret = QMessageBox.warning(self, "Clear Settings", 
                                "This action will clear all program settings, are you sure you want to do this?",
                                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        if ret == QMessageBox.StandardButton.Yes:
            items = [self.cmbViewerProfile.itemText(i) for i in range(self.cmbViewerProfile.count())]
            for item in items:
                print(item)
                QSettings("onvif-gui", item).clear()
                QSettings("onvif-gui", "gui").clear()
            QMessageBox.information(self, "Clear Settings",
                                    "All settings have been cleared, please re-start the application",
                                    QMessageBox.StandardButton.Ok)
    
    def btnTestClicked(self):
        print("btnTestClicked")




#!/usr/bin/env python3

import sys
import subprocess
from pathlib import Path

DIST_DIR = Path("dist")
print(DIST_DIR)

def python_tag():
    v = sys.version_info
    return f"cp{v.major}{v.minor}"

def run_pip(args):
    cmd = [sys.executable, "-m", "pip"] + args
    print(">>>", " ".join(cmd))
    subprocess.check_call(cmd)

def main():
    tag = python_tag()
    print(f"Detected Python tag: {tag}")

    wheels = list(DIST_DIR.glob("*.whl"))

    # Universal wheels (py3-none-any)
    universal = [w for w in wheels if "py3-none-any" in w.name]

    # ABI-specific wheels
    abi_specific = [w for w in wheels if f"-{tag}-" in w.name]

    if not abi_specific and not universal:
        print("No compatible wheels found.")
        sys.exit(1)

    # Separate onvif_gui
    gui_wheels = [w for w in universal + abi_specific if w.name.startswith("onvif_gui")]
    other_wheels = [w for w in universal + abi_specific if w not in gui_wheels]

    # Install non-GUI wheels first
    for wheel in sorted(other_wheels):
        run_pip(["install", str(wheel)])

    # Install GUI last
    for wheel in gui_wheels:
        run_pip(["install", str(wheel)])

    print("Installation complete.")

if __name__ == "__main__":
    main()



#/********************************************************************
# onvif-gui/onvif_gui/panels/camera/camerapanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from time import sleep
from PyQt6.QtWidgets import QPushButton, QGridLayout, QWidget, QSlider, \
    QListWidget, QTabWidget, QMessageBox, QMenu, QFileDialog
from PyQt6.QtGui import QAction
from PyQt6.QtCore import Qt, pyqtSignal, QObject, QTimer, QSettings
from . import NetworkTab, ImageTab, VideoTab, PTZTab, SystemTab, LoginDialog, \
    Session, Camera, CameraList, Snapshot
from loguru import logger
import libonvif as onvif
from pathlib import Path
import os
import subprocess
from datetime import datetime
from onvif_gui.enums import ProxyType, SnapshotAuth
import platform
import webbrowser
import requests
from requests.auth import HTTPDigestAuth
from urllib.parse import urlparse, parse_qs
import threading
import sys
import time

class CameraPanelSignals(QObject):
    fill = pyqtSignal(onvif.Data)
    login = pyqtSignal(onvif.Data)
    collapseSplitter = pyqtSignal()
    guiSync = pyqtSignal()

class CameraPanel(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.dlgLogin = LoginDialog(self)
        self.fillers = []
        self.sync_lock = False
        self.snapshot = Snapshot(mw)

        self.autoTimeSyncer = None
        self.enableAutoTimeSync(self.mw.settingsPanel.general.chkAutoTimeSync.isChecked())
       
        self.lstCamera = CameraList(mw)
        self.lstCamera.currentItemChanged.connect(self.onCurrentItemChanged)
        self.lstCamera.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.lstCamera.itemClicked.connect(self.onItemClicked)
        self.lstCamera.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.lstCamera.customContextMenuRequested.connect(self.showContextMenu)

        self.sldVolume = QSlider(Qt.Orientation.Horizontal)
        self.sldVolume.setValue(80)
        self.sldVolume.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.sldVolume.valueChanged.connect(self.sldVolumeChanged)
        self.sldVolume.setEnabled(False)

        self.btnStop = QPushButton()
        self.btnStop.setMinimumWidth(40)
        self.btnStop.setMaximumHeight(20)
        self.btnStop.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnStop.clicked.connect(self.btnStopClicked)

        self.btnRecord = QPushButton()
        self.btnRecord.setMinimumWidth(40)
        self.btnRecord.setMaximumHeight(20)
        self.btnRecord.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnRecord.clicked.connect(self.btnRecordClicked)

        self.btnMute = QPushButton()
        self.btnMute.setMinimumWidth(40)
        self.btnMute.setMaximumHeight(20)
        self.btnMute.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnMute.clicked.connect(self.btnMuteClicked)

        self.btnDiscover = QPushButton()
        self.btnDiscover.setMinimumWidth(40)
        self.btnDiscover.setMaximumHeight(20)
        self.btnDiscover.setStyleSheet(self.getButtonStyle("discover"))
        self.btnDiscover.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnDiscover.clicked.connect(self.btnDiscoverClicked)

        self.btnApply = QPushButton()
        self.btnApply.setMinimumWidth(40)
        self.btnApply.setMaximumHeight(20)
        self.btnApply.setStyleSheet(self.getButtonStyle("apply"))
        self.btnApply.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnApply.clicked.connect(self.btnApplyClicked)
        self.btnApply.setEnabled(False)
        
        self.btnSnapshot = QPushButton()
        self.btnSnapshot.setMinimumWidth(40)
        self.btnSnapshot.setMaximumHeight(20)
        self.btnSnapshot.setStyleSheet(self.getButtonStyle("snapshot"))
        self.btnSnapshot.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnSnapshot.clicked.connect(self.btnSnapshotClicked)
        self.btnSnapshot.setEnabled(False)        

        self.btnHelp = QPushButton()
        self.btnHelp.setMinimumWidth(40)
        self.btnHelp.setMaximumHeight(20)
        self.btnHelp.setStyleSheet(self.getButtonStyle("help"))
        self.btnHelp.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnHelp.clicked.connect(self.btnHelpClicked)
        
        self.btnStopAll = QPushButton()
        self.btnStopAll.setMinimumWidth(40)
        self.btnStopAll.setMaximumHeight(20)
        self.btnStopAll.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnStopAll.clicked.connect(self.btnStopAllClicked)

        self.btnHistory = QPushButton()
        self.btnHistory.setMinimumWidth(40)
        self.btnHistory.setMaximumHeight(20)
        self.btnHistory.setStyleSheet(self.getButtonStyle("history"))
        self.btnHistory.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnHistory.clicked.connect(self.btnHistoryClicked)

        self.btnFullScreen = QPushButton()
        self.btnFullScreen.setMinimumWidth(40)
        self.btnFullScreen.setMaximumHeight(20)
        self.btnFullScreen.setStyleSheet(self.getButtonStyle("full_screen"))
        self.btnFullScreen.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.btnFullScreen.clicked.connect(self.btnFullScreenClicked)

        self.tabOnvif = QTabWidget()
        self.tabOnvif.setUsesScrollButtons(False)
        self.tabVideo = VideoTab(self)
        self.tabImage = ImageTab(self)
        self.tabNetwork = NetworkTab(self)
        self.ptzTab = PTZTab(self)
        self.tabSystem = SystemTab(self)
        self.tabOnvif.addTab(self.tabVideo, "Media")
        self.tabOnvif.addTab(self.tabImage, "Image")
        self.tabOnvif.addTab(self.tabNetwork, "Network")
        self.tabOnvif.addTab(self.ptzTab, "PTZ")
        self.tabOnvif.addTab(self.tabSystem, "System")

        self.signals = CameraPanelSignals()
        self.signals.fill.connect(self.tabVideo.fill)
        self.signals.fill.connect(self.tabImage.fill)
        self.signals.fill.connect(self.tabNetwork.fill)
        self.signals.fill.connect(self.ptzTab.fill)
        self.signals.fill.connect(self.tabSystem.fill)
        self.signals.fill.connect(self.syncGUI)
        self.signals.login.connect(self.onShowLogin)
        self.signals.collapseSplitter.connect(self.mw.collapseSplitter)
        self.signals.guiSync.connect(self.syncGUI)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.lstCamera,     0, 0, 1, 6)
        lytMain.addWidget(self.tabOnvif,      1, 0, 1, 6)
        lytMain.addWidget(self.btnStopAll,    2, 0, 1, 1)
        lytMain.addWidget(self.btnHistory,    2, 1, 1, 1)
        lytMain.addWidget(self.btnSnapshot,   2, 2, 1, 1)
        lytMain.addWidget(self.btnFullScreen, 2, 3, 1, 1)
        lytMain.addWidget(self.btnHelp,       2, 4, 1, 1)
        lytMain.addWidget(self.btnStop,       3, 0, 1, 1)
        lytMain.addWidget(self.btnRecord,     3, 1, 1, 1)
        lytMain.addWidget(self.btnDiscover,   3, 2, 1, 1)
        lytMain.addWidget(self.btnApply,      3, 3, 1, 1)
        lytMain.addWidget(self.btnMute,       3, 4, 1, 1)
        lytMain.addWidget(self.sldVolume,     3, 5, 1, 1)
        lytMain.setColumnStretch(5, 10)
        lytMain.setRowStretch(0, 10)

        self.menu = QMenu("Context Menu", self)
        self.remove = QAction("Delete", self)
        self.rename = QAction("Rename", self)
        self.info = QAction("Info", self)
        self.password = QAction("Password", self)
        self.start = QAction("Start", self)
        self.stop = QAction("Stop", self)
        self.remove.triggered.connect(self.onMenuRemove)
        self.rename.triggered.connect(self.onMenuRename)
        self.info.triggered.connect(self.onMenuInfo)
        self.password.triggered.connect(self.onMenuPassword)
        self.start.triggered.connect(self.onMenuStart)
        self.stop.triggered.connect(self.onMenuStop)
        self.menu.addAction(self.remove)
        self.menu.addAction(self.rename)
        self.menu.addAction(self.info)
        self.menu.addAction(self.password)
        self.menu.addAction(self.start)
        self.menu.addAction(self.stop)

        self.syncGUI()
        self.setTabsEnabled(False)
        self.sessions = []
        self.closing = False

    def showContextMenu(self, pos):
        index = self.lstCamera.indexAt(pos)
        if index.isValid():
            self.menu.exec(self.mapToGlobal(pos))

    def onMenuRemove(self):
        self.lstCamera.remove()
    
    def onMenuRename(self):
        self.lstCamera.rename()

    def onMenuInfo(self):
        self.lstCamera.info()

    def onMenuPassword(self):
        self.lstCamera.password()

    def onMenuStart(self):
        self.lstCamera.startCamera()

    def onMenuStop(self):
        self.lstCamera.stopCamera()

    def btnDiscoverClicked(self):
        if self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
            if self.mw.client:
                self.mw.client.transmit(bytearray("GET CAMERAS\r\n", 'utf-8'))
            return
        
        if self.mw.settingsPanel.discover.radDiscover.isChecked():
            logger.debug("Using broadcast discovery")
            interfaces = []
            self.sessions.clear()

            if self.mw.settingsPanel.discover.chkScanAllNetworks.isChecked():
                for i in range(self.mw.settingsPanel.discover.cmbInterfaces.count()):
                    interfaces.append(self.mw.settingsPanel.discover.cmbInterfaces.itemText(i))
            else:
                interfaces.append(self.mw.settingsPanel.discover.cmbInterfaces.currentText())

            for interface in interfaces:
                session = Session(self, interface)
                session.start()
                self.sessions.append(session)
                self.btnDiscover.setEnabled(False)
                if len(interfaces) > 1:
                    sleep(1)
        else:
            logger.debug("Using cached camera addresses for discovery")
            self.fillers.clear()
            tmp = self.mw.settings.value(self.mw.settingsPanel.discover.cameraListKey)
            if tmp:
                numbers = tmp.strip().split("\n")
                for serial_number in numbers:
                    key = f'{serial_number}/XAddrs'
                    xaddrs = self.mw.settings.value(key)
                    alias = self.mw.settings.value(f'{serial_number}/Alias')
                    data = onvif.Data()
                    data.errorCallback = self.errorCallback
                    data.infoCallback = self.infoCallback
                    data.setSetting = self.mw.settings.setValue
                    data.getSetting = self.mw.settings.value
                    data.getData = self.getData
                    data.getCredential = self.getCredential
                    data.setXAddrs(xaddrs)
                    data.alias = alias
                    data.setCameraName(alias)
                    data.setDeviceService("POST /onvif/device_service HTTP/1.1\r\n")
                    self.fillers.append(data)
                    data.startManualFill()

    def errorCallback(self, msg):
        try:
            self.mw.signals.error.emit(msg)
        except Exception as ex:
            logger.error("camera panel error callback exception: {ex}")

    def infoCallback(self, msg):
        try:
            if msg.startswith("Set System Date and Time Error"):
                logger.error(msg)
            else:
                logger.debug(msg)
        except Exception as ex:
            logger.error("camera panel info callback exception: {ex}")

    def discoveryTimeout(self):
        self.setEnabled(True)
        self.btnDiscover.setEnabled(True)

    def discovered(self):
        finished = True
        for session in self.sessions:
            if session.active:
                finished = False
                break

        if finished:
            self.sessions.clear()
            self.btnDiscover.setEnabled(True)
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.SERVER:
                self.mw.settingsPanel.proxy.setMediaMTXProxies()

    def getCredential(self, onvif_data):
        if not onvif_data: return
        
        if self.getCameraByXAddrs(onvif_data.xaddrs()) and not len(self.fillers):
            onvif_data.cancelled = True
            return onvif_data
                
        alternateUsername = self.mw.settings.value(f'{onvif_data.xaddrs()}/alternateUsername', None)
        alternatePassword = self.mw.settings.value(f'{onvif_data.xaddrs()}/alternatePassword', None)

        if (len(self.mw.settingsPanel.general.txtPassword.text()) or alternatePassword) and len(onvif_data.last_error()) == 0:
            if alternateUsername:
                onvif_data.setUsername(alternateUsername)
            else:
                onvif_data.setUsername(self.mw.settingsPanel.general.txtUsername.text())
            if alternatePassword:
                onvif_data.setPassword(alternatePassword)
            else:
                onvif_data.setPassword(self.mw.settingsPanel.general.txtPassword.text())
        else:
            if onvif_data.last_error().startswith("Network error, unable to connect"):
                logger.debug(f'Unable to connect with {onvif_data.xaddrs()}')
                onvif_data.cancelled = True
                self.mw.signals.error.emit(f'Unable to connect with {onvif_data.xaddrs()}')
            else:
                while self.dlgLogin.active:
                    sleep(0.01)

                self.dlgLogin.active = True
                self.signals.login.emit(onvif_data)
                while self.dlgLogin.active:
                    sleep(0.01)

        return onvif_data
    
    def onShowLogin(self, onvif_data):
        self.dlgLogin.exec(onvif_data)

    def getProxyData(self, onvif_data):
        if not onvif_data: return
        
        onvif_data.getProxyURI = self.mw.getProxyURI

        alias = self.mw.settings.value(f'{onvif_data.serial_number()}/Alias')
        if not alias:
            name = onvif_data.camera_name()
            if len(name):
                alias = name
                self.mw.settings.setValue(f'{onvif_data.serial_number()}/Alias', name)
            else:
                alias = onvif_data.host()
        onvif_data.alias = alias

        if not self.getCameraBySerialNumber(onvif_data.serial_number()):

            self.mw.alarm_ordinals[len(self.mw.alarm_ordinals)] = onvif_data.serial_number()

            add_camera = True
            if self.mw.settingsPanel.discover.radCached.isChecked() and self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
                if tmp := self.mw.settings.value(self.mw.settingsPanel.discover.cameraListKey):
                    numbers = tmp.strip().split("\n")
                    if onvif_data.serial_number() not in numbers:
                        add_camera = False

            if add_camera:
                camera = Camera(onvif_data, self.mw)
                camera.setIconIdle()
                camera.dimForeground()
                self.mw.addCameraProxy(camera)
                self.lstCamera.addItem(camera)
                self.lstCamera.sortItems()
                camera.setDisplayProfile(camera.getDisplayProfileSetting())
                logger.debug(f'Discovery completed for Camera: {onvif_data.alias}, Stream URI: {onvif_data.stream_uri()}, xaddrs: {onvif_data.xaddrs()}')

                self.filled(onvif_data)

    def getData(self, onvif_data):
        if not onvif_data: return
        
        if onvif_data.last_error().startswith("Error initializing camera data during manual fill:"):
            logger.debug(onvif_data.last_error())
            return

        onvif_data.filled = self.filled
        onvif_data.infoCallback = self.infoCallback
        onvif_data.errorCallback = self.errorCallback

        alias = self.mw.settings.value(f'{onvif_data.serial_number()}/Alias')
        if not alias:
            name = onvif_data.camera_name()
            if len(name):
                alias = name
                self.mw.settings.setValue(f'{onvif_data.serial_number()}/Alias', name)
            else:
                alias = onvif_data.host()
        onvif_data.alias = alias

        if existing := self.getCameraBySerialNumber(onvif_data.serial_number()):
            synchronizeTime = self.mw.settingsPanel.general.chkAutoTimeSync.isChecked()
            if not self.closing:
                existing.onvif_data.setXAddrs(onvif_data.xaddrs())
                for profile in existing.profiles:
                    profile.setXAddrs(onvif_data.xaddrs())
                existing.onvif_data.startFill(synchronizeTime)
        else:
            camera = Camera(onvif_data, self.mw)
            camera.setIconIdle()
            camera.dimForeground()
            self.mw.addCameraProxy(camera)
            
            self.lstCamera.addItem(camera)
            self.lstCamera.sortItems()
            camera.setDisplayProfile(camera.getDisplayProfileSetting())
            self.saveCameraList()
            logger.debug(f'Discovery completed for Camera: {onvif_data.alias}, Stream URI: {onvif_data.stream_uri()}, xaddrs: {onvif_data.xaddrs()}, {onvif_data.camera_name()}')

            synchronizeTime = self.mw.settingsPanel.general.chkAutoTimeSync.isChecked()
            if not self.closing:
                onvif_data.startFill(synchronizeTime)

    def filled(self, onvif_data):
        if not onvif_data: return
        
        if camera := self.getCamera(onvif_data.uri()):
            camera.restoreForeground()
            key = f'{camera.serial_number()}/XAddrs'
            self.mw.settings.setValue(key, camera.xaddrs())

            if camera.manual_fill:
                camera.assignData(onvif_data)
                self.mw.addCameraProxy(camera)
                camera.setDisplayProfile(camera.getDisplayProfileSetting())

            if self.lstCamera is not None:
                current_camera = self.getCurrentCamera()
                if current_camera:
                    if current_camera.xaddrs() == onvif_data.xaddrs():
                        self.signals.fill.emit(onvif_data)
                        self.setEnabled(True)
                        self.setTabsEnabled(True)

            camera.filled = True
            if self.allCamerasFilled():
                self.signals.guiSync.emit()
            if self.mw.settingsPanel.proxy.proxyType == ProxyType.SERVER and \
                    self.mw.settingsPanel.discover.radCached.isChecked() and \
                    self.allCamerasFilled():
                
                self.mw.settingsPanel.proxy.setMediaMTXProxies()

            # auto start after fill, recording needs onvif frame rate
            if self.mw.settingsPanel.discover.chkAutoStart.isChecked():
                if not camera.isRunning():
                    while not self.mw.isVisible():
                        sleep(0.1)
                    self.lstCamera.itemClicked.emit(camera)
                    self.lstCamera.itemDoubleClicked.emit(camera)
                    sleep(0.1)

        if len(onvif_data.last_error()):
            logger.debug(f'Error from {onvif_data.alias} : {onvif_data.last_error()}')

    def saveCameraList(self):
        serial_numbers = ""
        cameras = [self.mw.cameraPanel.lstCamera.item(x) for x in range(self.mw.cameraPanel.lstCamera.count())]
        for camera in cameras:
            serial_numbers += camera.serial_number() + "\n"
        self.mw.settings.setValue(self.mw.settingsPanel.discover.cameraListKey, serial_numbers)

    def onCurrentItemChanged(self, current, previous):
        if current:
            if self.mw.pm.getPlayer(current.uri()):
                self.mw.glWidget.focused_uri = current.uri()
            else:
                self.mw.glWidget.focused_uri = None
            self.signals.fill.emit(current.onvif_data)
            self.syncGUI()

    def onItemClicked(self, camera):
        if self.mw.videoConfigure:
            self.mw.videoConfigure.setCamera(camera)

    def onItemDoubleClicked(self, camera):
        if not camera: return
        profiles = self.mw.pm.getStreamPairProfiles(camera.uri())
        players = self.mw.pm.getStreamPairPlayers(camera.uri())
        timers = self.mw.pm.getStreamPairTimers(camera.uri())

        activeTimer = False
        for timer in timers:
            if timer.isActive():
                activeTimer = True

        if activeTimer:
            for timer in timers:
                self.mw.signals.stopReconnect.emit(timer.uri)
            for player in players:
                player.requestShutdown()
            camera.setIconIdle()
        else:
            if len(players):
                if player := self.mw.cameraPanel.getCurrentPlayer():
                    if profile := camera.getRecordProfile():
                        self.mw.openFocusWindow()
                        count = 0
                        while not self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                            time.sleep(0.01)
                            count += 1
                            if count > 200:
                                logger.error("timeout error opening focus window")
                                break
                        if camera := self.mw.focus_window.cameraPanel.getCamera(profile.uri()):
                            self.mw.focus_window.cameraPanel.onItemDoubleClicked(camera)
            else:
                for i, profile in enumerate(profiles):
                    if i == 0:
                        profile.setHidden(False)
                        self.mw.playMedia(profile.uri())
                    else:
                        if camera.displayProfileIndex() != camera.recordProfileIndex():
                            profile.setHidden(True)
                            self.mw.playMedia(profile.uri())

        self.syncGUI()

    def setTabsEnabled(self, enabled):
        self.tabVideo.setEnabled(enabled)
        self.tabImage.setEnabled(enabled)
        self.tabNetwork.setEnabled(enabled)
        self.ptzTab.setEnabled(enabled)
        self.tabSystem.setEnabled(enabled)

    def btnApplyClicked(self):
        if camera := self.getCurrentCamera():
            self.btnApply.setEnabled(False)
            self.tabVideo.update(camera.onvif_data)
            self.tabImage.update(camera.onvif_data)
            self.tabNetwork.update(camera.onvif_data)

    def onEdit(self):
        if camera := self.getCurrentCamera():
            if self.tabVideo.edited(camera.onvif_data) or \
                    self.tabImage.edited(camera.onvif_data) or \
                    self.tabNetwork.edited(camera.onvif_data):
                self.btnApply.setEnabled(True)
            else:
                self.btnApply.setEnabled(False)

    def sldVolumeChanged(self, value):
        if player := self.getCurrentPlayer():
            player.setVolume(value)
        if camera := self.getCurrentCamera():
            camera.setVolume(value)

    def btnMuteClicked(self):
        player = self.getCurrentPlayer()
        if player:
            player.setMute(not player.isMuted())
            camera = self.getCurrentCamera()
            if camera:
                camera.setMute(player.isMuted())
        else:
            camera = self.getCurrentCamera()
            if camera:
                camera.setMute(not camera.mute)
        self.syncGUI()

    def btnRecordClicked(self):
        player = self.getCurrentPlayer()
        camera = self.getCurrentCamera()
        if camera.displayProfileIndex() != camera.recordProfileIndex():
            recordProfile = camera.getRecordProfile()
            if recordProfile:
                record_uri = recordProfile.uri()
                player = self.mw.pm.getPlayer(record_uri)

        if player:
            if player.isRecording():
                player.output_file_start_time = None
                player.toggleRecording("")
                self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                if camera:
                    camera.manual_recording = False
            else:
                #d = self.mw.settingsPanel.storage.dirArchive.txtDirectory.text()
                #if self.mw.settingsPanel.storage.chkManageDiskUsage.isChecked():
                #    self.mw.diskManager.manageDirectory(d)
                #else:
                #    self.mw.settingsPanel.storage.signals.updateDiskUsage.emit()
                if filename := player.getOutputFilename():
                    player.toggleRecording(filename)
                    if camera:
                        camera.manual_recording = True

        self.syncGUI()

    def btnHistoryClicked(self):
        try:
            reader_settings = QSettings("onvif-gui", "Reader")
            reader_settings.setValue("filePanel/hideCameraPanel", 1)
            main_file = Path(__file__).parent.parent.parent / "main.py"
            #subprocess.Popen(["python", str(main_file), "--profile", "reader"], env=os.environ.copy(), start_new_session=True, shell=True)
            if platform.system() == "Windows":
                subprocess.Popen([sys.executable, str(main_file), "--profile", "Reader"], env=os.environ.copy(), start_new_session=True, shell=True)
            else:
                subprocess.Popen([sys.executable, str(main_file), "--profile", "Reader"], env=os.environ.copy(), start_new_session=True)
        except Exception as ex:
            logger.error(f'Error starting file browser: {ex}')
            self.mw.signals.error.emit(f'Error starting file browser: {ex}')

    def btnStopClicked(self):
        if camera := self.getCurrentCamera():
            #profiles = self.mw.pm.getStreamPairProfiles(camera.uri())
            players = self.mw.pm.getStreamPairPlayers(camera.uri())
            timers = self.mw.pm.getStreamPairTimers(camera.uri())

            activeTimer = False
            for timer in timers:
                if timer.isActive():
                    activeTimer = True

            if activeTimer:
                for timer in timers:
                    self.mw.signals.stopReconnect.emit(timer.uri)
                for player in players:
                    player.requestShutdown()
                camera.setIconIdle()
                self.syncGUI()
                return

            if len(players):
                for player in players:
                    player.requestShutdown()
                camera.setIconIdle()
                self.syncGUI()
                return
            else:
                self.onItemDoubleClicked(camera)
                self.syncGUI()

    def btnHelpClicked(self):
        result = webbrowser.get().open("https://github.com/sr99622/libonvif#readme-ov-file")
        if not result:
            webbrowser.get().open("https://github.com/sr99622/libonvif")

    def btnSnapshotClicked(self):
        if player := self.getCurrentPlayer():
            root = self.mw.settingsPanel.storage.dirPictures.txtDirectory.text() + "/" + self.getCamera(player.uri).text()
            Path(root).mkdir(parents=True, exist_ok=True)
            filename = '{0:%Y%m%d%H%M%S.jpg}'.format(datetime.now())
            filename = str(root + "/" + filename)

            if self.mw.settingsPanel.general.chkSnapshotDlg.isChecked():
                if platform.system() == "Linux":
                    filename = QFileDialog.getSaveFileName(self, "Save File As", filename, options=QFileDialog.Option.DontUseNativeDialog)[0]
                else:
                    filename = QFileDialog.getSaveFileName(self, "Save File As", filename)[0]

            if len(Path(filename).stem):
                answer = QMessageBox.StandardButton.Yes
                if not filename.endswith(".jpg"):
                    filename += ".jpg"
                    if Path(filename).is_file():
                        answer = QMessageBox.question(self.mw, "File Exists", "You are about to overwrite an existing file, are you sure you wnat to do this?")
                if answer == QMessageBox.StandardButton.Yes:
                    if camera := self.getCamera(player.uri):
                        profile = camera.getRecordProfile()
                        if not profile:
                            profile = camera.getProfile(player.uri)

                        thread = threading.Thread(target=self.snapshot, args=(profile, filename, camera, player))
                        thread.start()

    def btnStopAllClicked(self):
        if len(self.mw.pm.players):
            self.mw.closeAllStreams()
        else:
            self.mw.startAllCameras()
        self.syncGUI()

    def btnFullScreenClicked(self):
        if self.mw.isFullScreen():
            self.mw.showNormal()
            self.btnFullScreen.setStyleSheet(self.getButtonStyle("full_screen"))
        else:
            self.mw.showFullScreen()
            self.btnFullScreen.setStyleSheet(self.getButtonStyle("normal"))
       
    def onMediaStarted(self, uri):
        if self.mw.tabVisible:
            if self.mw.glWidget.focused_uri is None:
                if camera := self.getCurrentCamera():
                    self.mw.glWidget.focused_uri = camera.uri()

        #self.tabVideo.btnSnapshot.setEnabled(True)
        
        self.syncGUI()

    def onMediaStopped(self, uri):
        if camera := self.getCamera(uri):
            camera.setIconIdle()
            if profile := camera.getProfile(camera.uri()):
                if profile.getAnalyzeVideo():
                    if self.mw.videoWorker:
                        self.mw.videoWorker(None, None)
                if profile.getAnalyzeAudio():
                    if self.mw.audioWorker:
                        self.mw.audioWorker(None, None)
        #self.tabVideo.btnSnapshot.setEnabled(False)
        self.syncGUI()

    def syncGUI(self):
        while (self.sync_lock):
            sleep(0.001)
        self.sync_lock = True
        
        if camera := self.getCurrentCamera():
            self.btnStop.setEnabled(True)
            if player := self.mw.pm.getPlayer(camera.uri()):
                self.btnStop.setStyleSheet(self.getButtonStyle("stop"))
                self.btnSnapshot.setEnabled(True)
                #self.tabVideo.btnSnapshot.setEnabled(True)

                if ps := player.systemTabSettings():
                    self.btnRecord.setEnabled(not (ps.record_enable and ps.record_always))

                if player.hasAudio() and not player.disable_audio:
                    self.btnMute.setEnabled(True)
                    self.sldVolume.setEnabled(True)
                    self.sldVolume.setValue(camera.volume)
                    if camera.mute:
                        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
                    else:
                        self.btnMute.setStyleSheet(self.getButtonStyle("audio"))
                else:
                    self.btnMute.setEnabled(False)
                    self.sldVolume.setEnabled(False)

                self.btnRecord.setEnabled(True)
                if camera.isRecording():
                    self.btnRecord.setStyleSheet(self.getButtonStyle("recording"))
                    record_always = player.systemTabSettings().record_always if player.systemTabSettings() else False
                    record_alarm = player.systemTabSettings().record_alarm if player.systemTabSettings() else False
                    record_enable = player.systemTabSettings().record_enable if player.systemTabSettings() else False
                    if record_enable and ((camera.isAlarming() and record_alarm) or record_always):
                        self.btnRecord.setEnabled(False)
                else:
                    self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
            else:
                reconnecting = False
                #self.tabVideo.btnSnapshot.setEnabled(False)
                timers = self.mw.pm.getStreamPairTimers(camera.uri())
                for timer in timers:
                    if timer.isActive():
                        reconnecting = True
                
                if reconnecting:
                    self.btnStop.setStyleSheet(self.getButtonStyle("stop"))
                else:
                    self.btnStop.setStyleSheet(self.getButtonStyle("play"))
                    self.setTabsEnabled(True)

                if profile := camera.getProfile(camera.uri()):
                    if camera.mute:
                        self.btnMute.setStyleSheet(self.getButtonStyle("mute"))
                    else:
                        self.btnMute.setStyleSheet(self.getButtonStyle("audio"))

                    if profile.audio_bitrate() and not profile.getDisableAudio():
                        self.btnMute.setEnabled(True)
                        self.sldVolume.setEnabled(True)
                    else:
                        self.btnMute.setEnabled(False)
                        self.sldVolume.setEnabled(False)

                self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
                self.btnRecord.setEnabled(False)
                self.btnSnapshot.setEnabled(False)
        else:
            self.sldVolume.setEnabled(False)
            self.btnMute.setStyleSheet(self.getButtonStyle("audio"))
            self.btnMute.setEnabled(False)
            self.btnRecord.setStyleSheet(self.getButtonStyle("record"))
            self.btnRecord.setEnabled(False)
            self.btnStop.setStyleSheet(self.getButtonStyle("play"))
            self.btnStop.setEnabled(False)
            self.btnSnapshot.setEnabled(False)

        if self.lstCamera.count():
            self.btnStopAll.setEnabled(True) 
        else:
            self.btnStopAll.setEnabled(False)
        if len(self.mw.pm.players):
            self.btnStopAll.setStyleSheet(self.getButtonStyle("stop_all"))
        else:
            self.btnStopAll.setStyleSheet(self.getButtonStyle("play_all"))

        self.sync_lock = False


    def getButtonStyle(self, name):
        strStyle = "QPushButton { image : url(image:%1.png); } \
                    QPushButton:hover { image : url(image:%1_hi.png); } \
                    QPushButton:pressed { image : url(image:%1_lo.png); } \
                    QPushButton:disabled { image : url(image:%1_lo.png); }"
        strStyle = strStyle.replace("%1", name)
        return strStyle

    def getCurrentPlayer(self):
        result = None
        if self.lstCamera:
            camera = self.getCurrentCamera()
            if camera:
                result = self.mw.pm.getPlayer(camera.uri())
        return result

    def getCamera(self, uri):
        if not uri: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            for profile in camera.profiles:
                if profile.uri() == uri:
                    return camera

        return None
    
    def getCameraByName(self, name):
        if not name: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.name() == name:
                return camera
                
        return None

    
    def getCameraBySerialNumber(self, serial_number):
        if not serial_number: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.serial_number() == serial_number:
                return camera
                
        return None
    
    def getCameraByXAddrs(self, xaddrs):
        if not xaddrs: return None
        if not self.lstCamera: return None

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if camera.xaddrs() == xaddrs:
                return camera
            
        return None
    
    def getProfile(self, uri):
        result = None
        if camera := self.getCamera(uri):
            result = camera.getProfile(uri)
        return result
    
    def getCurrentProfile(self):
        result = None
        if camera := self.getCurrentCamera():
            result = camera.getProfile(camera.uri())
        return result
    
    def getCurrentCamera(self):
        result = None
        if self.lstCamera:
            result = self.lstCamera.currentItem()
        return result
    
    def setCurrentCamera(self, uri):
        if camera := self.getCamera(uri):
            self.lstCamera.setCurrentItem(camera)
            self.signals.fill.emit(camera.onvif_data)
            self.syncGUI()

            if self.mw.videoConfigure:
                    self.mw.videoConfigure.setCamera(camera)

            if self.mw.audioConfigure:
                self.mw.audioConfigure.setCamera(camera)

    def enableAutoTimeSync(self, state):
        AUTO_TIME_SYNC_INTERVAL = 3600000
        if int(state) == 0:
            logger.debug("Auto time sync has been turned off")
            if self.autoTimeSyncer:
                self.autoTimeSyncer.stop()
        else:
            logger.debug("Auto time sync has been turned on")
            if not self.autoTimeSyncer:
                self.autoTimeSyncer = QTimer()
                self.autoTimeSyncer.setInterval(AUTO_TIME_SYNC_INTERVAL)
                self.autoTimeSyncer.timeout.connect(self.timeSync)
                self.autoTimeSyncer.start()
    
    def timeSync(self):
        if not self.lstCamera: return
        logger.debug("Synchronizing camera times")
        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            camera.onvif_data.startUpdateTime()

    def activeSessions(self):
        for session in self.sessions:
            if session.active:
                return True
            
        return False

    def closeEvent(self):
        self.closing = True

        if self.lstCamera:
            cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
            for camera in cameras:
                camera.onvif_data.filled = None

        if self.activeSessions():
            for session in self.sessions:
                session.abort = True

            sleep(1)
            
            waiting = True
            count = 0
            while waiting:
                count += 1
                tmp = False
                for session in self.sessions:
                    if session.active:
                        tmp = True
                        break
                    sleep(0.1)
                waiting = tmp
                if count > 10:
                    break

    def allCamerasFilled(self):
        if not self.lstCamera: return True

        cameras = [self.lstCamera.item(x) for x in range(self.lstCamera.count())]
        for camera in cameras:
            if not camera.filled:
                return False

        return True


#********************************************************************
# libonvif/assets/scripts/install-onvif-gui.py
#
# Copyright (c) 2025  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

import os
import sys
import shutil
import platform
import argparse

ERROR = 1
SUCCESS = 0

INSTALL_SOURCE = "pypi"


#INSTALL_SOURCE = "smb"
SMB_SERVER = "//10.1.1.143/share"
SMB_DOMAIN = "WORKGROUP"
SMB_USERNAME = "username"
SMB_PASSWORD = "password"
SMB_MOUNT = "smb_share"
SMB_FOLDER = "home/safe/install/3.0.10/linux"

# credential = f"\"domain={domain},username={user},password={passwd}\""
# f'sudo mount -t cifs -o {credential} {host} {target}'

class Install():
    def __init__(self):
        self.status = SUCCESS
        self.venv = None

    def run_command(self, command, env=None, exit_on_error=False, silent=True):
        if env is None:
            env = ""
            env_dict = os.environ.copy()
            for key in env_dict:
                env += f'{key}={env_dict[key]}\n'

        stdout_r, stdout_w = os.pipe()
        stderr_r, stderr_w = os.pipe()
        pid = os.fork()

        if pid == 0:
            os.close(stdout_r)  
            os.close(stderr_r)
            os.dup2(stdout_w, 1)
            os.dup2(stderr_w, 2)
            os.close(stdout_w)
            os.close(stderr_w)
            os.execlp("bash", "bash", "-c", command, env)
        else:
            os.close(stdout_w)
            os.close(stderr_w)
            with os.fdopen(stdout_r) as stdout_pipe, os.fdopen(stderr_r) as stderr_pipe:
                output = stdout_pipe.read().strip()
                error_output = stderr_pipe.read().strip()
            pid, self.status = os.waitpid(pid, 0)

            if os.WIFEXITED(self.status) and os.WEXITSTATUS(self.status) != 0:
                if not silent:
                    print(f"Command failed: {command}", file=sys.stderr)
                    print(f"Error message: {error_output}", file=sys.stderr)
                if exit_on_error:
                    sys.exit(os.WEXITSTATUS(self.status))
            return output + error_output

    def install_package(self, pkg_mgr, name):
        print(f"installing {name} using package manager {pkg_mgr}")

        if pkg_mgr == "apt":
            format = "-Wf'${db:Status-abbrev}'"
            self.run_command(f"dpkg-query {format} {name} | grep -q '^i'")
            if self.status:
                print(f"Existing package was not found, installing {name}")
                print(self.run_command(f'sudo apt install -y {name}'))
                if self.status:
                    print(f"An error occurred installing the package: {name}, error code: {self.status}")
                    return ERROR
            else:
                print("Found existing package")
        
        if pkg_mgr == "dnf":
            print(self.run_command(f'sudo dnf install -y {name}'))
            if self.status:
                print(f"An error occurred installing the package: {name}, error code: {self.status}")
                return ERROR
        
        if pkg_mgr == "pacman":
            print("The installation program does not install packages on systems that use pacman")

    def check_python(self, pkg_mgr):
        major, minor, micro = platform.python_version_tuple()
        print(f"The Python version for this installation is {major}.{minor}.{micro}")       
        supported = True 
        if int(major) < 3:
            print("Unsupported Python version, the minimum is 3.10")
            supported = False
        if int(minor) < 10:
            print("Unsupported Python version, the minimum is 3.10")
            supported = False

        if not supported:
            kernel_supported = False
            kernel_version = platform.release()
            print(f"kernel version {kernel_version}")
            if len(kernel_version) > 1:
                try:
                    kernel_major = int(kernel_version.split(".")[0])
                    kernel_minor = int(kernel_version.split(".")[1])
                    if kernel_major * 100 + kernel_minor >= 504:
                        kernel_supported = True
                except Exception as ex:
                    kernel_supported = False

            if not kernel_supported:
                print(f"Unsupported kernel {kernel_version}, the minimum required kernel version is 5.4")
                return ERROR
            
            print("Attempting to find a compatible python version")
            if valid_existing_python_version := self.check_for_alt_python_versions():
                print(f"\nThere is an existing supported version {valid_existing_python_version} on this machine.\nOnvif GUI will be installed using this version, please wait ...")
                print(self.run_command(f"{valid_existing_python_version} install-onvif-gui.py"))
            else:
                if response := input("Would you like to safely install an alternate Python version (3.12) for running Onvif GUI N/y? "):
                    if response.lower() == "y":
                        if self.build_python(pkg_mgr):
                            print("\n\nPython3.12 installation SUCCESSFUL. Press the enter key to continue ...")
                        else:
                            print("\n\n An error occurred attempting to build Python 3.12")
                        print(self.run_command('sudo rm -rf Python-3.12.9*'))
                        print(self.run_command("python3.12 install-onvif-gui.py"))
                    else:
                        print("Onvif GUI installation has been cancelled")
            return ERROR

    def build_python(self, pkg_mgr):
        
        if pkg_mgr == "apt":
            packages = ["curl", "libssl-dev", "zlib1g-dev", "libbz2-dev", "build-essential",
                        "libreadline-dev", "libsqlite3-dev", "wget", "llvm", 
                        "libncurses5-dev", "libncursesw5-dev", "xz-utils", "tk-dev", 
                        "libffi-dev", "liblzma-dev", "python-openssl", "git", "pkg-config"]
        
        elif pkg_mgr == "dnf":
            packages = ["gcc", "openssl-devel", "bzip2-devel", "libffi-devel", "wget", "tar", 
                        "make", "zlib-devel", "sqlite-devel", "xz-devel", "ncurses-devel", 
                        "readline-devel", "tk-devel", "libuuid-devel"]

        else:
            print("Unsupported package manager for Python compile")
            return False

        for package in packages:
            print(f"package: {package}")
            self.install_package(pkg_mgr, package)

        print(f"{self.run_command('wget https://www.python.org/ftp/python/3.12.9/Python-3.12.9.tgz')}")
        if self.status:
            print("An error occurred downloading the Python source code")
            return False
        
        print("Python 3.12 will be compiled from source. This will take some time ...")
        command = "tar -xf Python-3.12.9.tgz && cd Python-3.12.9 && ./configure --enable-optimizations && make -j$(nproc) && sudo make altinstall"
        output = self.run_command(command)
        if self.status:
            print(f'An error occurred during installation for Python3.12\v{output}')
            return False
        
        return True

    def check_for_alt_python_versions(self):
        self.run_command('python3.10 --version', silent=True)
        if not self.status:
            return "python3.10"
        self.run_command('python3.11 --version', silent=True)
        if not self.status:
            return "python3.11"
        self.run_command('python3.12 --version', silent=True)
        if not self.status:
            return "python3.12"
        self.run_command('python3.13 --version', silent=True)
        if not self.status:
            return "python3.13"
        return None

    def create_venv(self, venv, pkg_mgr):
        print(f"Creating virtual environment at {venv}")
        major, minor, _ = platform.python_version_tuple()
        print(self.run_command(f'python{major}.{minor} -m venv {venv}'))
        if self.status:
            print("Missing venv module, will attempt to install")
            pkg_name = f'python{major}.{minor}-venv'
            if self.install_package(pkg_mgr, pkg_name):
                return ERROR
            print(self.run_command(f'python{major}.{minor} -m venv {venv}'))
            if self.status:
                return ERROR
        self.venv = venv

    def check_display_protocol(self, pkg_mgr):
        # This function is allowed to fail without returning ERROR
        print("Checking display protocol")
        
        if protocol := os.environ.get("XDG_SESSION_TYPE"):

            if protocol.casefold() == "wayland".casefold():
                print("Display protocol is wayland, no further configuration required")
                return

            if protocol.casefold() == "x11".casefold():
                if pkg_mgr == "apt":
                    self.install_package(pkg_mgr, "libxcb-cursor-dev")
                if pkg_mgr == "dnf":
                    ...
                if pkg_mgr == "pacman":
                    ...

            else:
                print(f"Unsupported display protocol: {protocol}")
        else:
            print(f"Unable to determine display protocol")
        
    def find_package_manager(self):
        self.run_command("apt list --installed")
        if not self.status:
            print("The package manager for this platform is apt")
            print(self.run_command('sudo apt update'))
            return "apt"
        self.run_command('dnf list')
        if not self.status:
            print("The package manager for this platform is dnf")
            #print(self.run_command('sudo dnf update'))
            return "dnf"
        self.run_command("pacman -Q")
        if not self.status:
            print("The package manager for this platform is pacman")
            #print(self.run_command('sudo pacman -Syy'))
            return "pacman"
        
        print("Unable to determine the package manager for this platform")
        return None
    
    def mount_smb(self, host, domain, user, passwd, target):
        try:
            os.makedirs(target, exist_ok=True)
        except Exception as ex:
            print(f"Error creating mount point: {ex}")
            return ERROR

        credential = f"\"domain={domain},username={user},password={passwd}\""

        print(self.run_command(f'sudo mount -t cifs -o {credential} {host} {target}'))
        if self.status:
            print(f"An error occurred while attempting to mount {host}")
            return ERROR
        
    def umount_smb(self, target):
        print(self.run_command(f'sudo umount {target}'))
        if self.status:
            print(f"An error occured during umount from {target}")
            return ERROR
        try:
            os.rmdir(target)
        except Exception as ex:
            print(f'An error occured while removing mount point {target}')
            return ERROR
        
    def install_onvif_gui(self, venv):
        print(self.run_command(f'source {venv}/bin/activate && pip install --upgrade pip && pip install onvif-gui'))
        if self.status:
            print("ERROR: Unable to properly install onvif-gui")
        return self.status
    
    def install_onvif_gui_local(self, venv, pkg_mgr):
        pkg_name = "cifs-utils"
        if self.install_package(pkg_mgr, pkg_name):
            return ERROR

        if self.mount_smb(SMB_SERVER, SMB_DOMAIN, SMB_USERNAME, SMB_PASSWORD, SMB_MOUNT):
            return ERROR
        
        major, minor, micro = platform.python_version_tuple()
        VER = f'{major}{minor}'
        dir = f'{SMB_MOUNT}/{SMB_FOLDER}'

        commands = [f"source {venv}/bin/activate && pip install --disable-pip-version-check {dir}/avio*{VER}*whl",
                    f"source {venv}/bin/activate && pip install --disable-pip-version-check {dir}/libonvif*{VER}*whl",
                    f"source {venv}/bin/activate && pip install --disable-pip-version-check {dir}/kankakee*{VER}*whl",
                    f"source {venv}/bin/activate && pip install --disable-pip-version-check {dir}/onvif*whl"]
        
        for command in commands:
            print(self.run_command(command))
            if self.status:
                print("An error occuring during pip install onvif-gui")
                return ERROR
            
        if self.umount_smb(SMB_MOUNT):
            return ERROR
        
    def check_for_nvidia(self):
        print(f"venv: {self.venv}")
        self.run_command('nvidia-smi')
        if self.status:
            print("Did not find NVIDIA drivers, installing CPU only pytorch")
            print(self.run_command(f'source {self.venv}/bin/activate && pip install --disable-pip-version-check torch torchvision --index-url https://download.pytorch.org/whl/cpu'))
        else:
            print("Found NVIDIA drivers, installing GPU version of pytorch, this may take a few minutes")
            print(self.run_command(f'source {self.venv}/bin/activate && pip install --disable-pip-version-check torch torchvision'))
        print(self.run_command(f'source {self.venv}/bin/activate && pip install --disable-pip-version-check openvino'))

    def rcwec(self, command, show=True):
        # run command with error check
        output = self.run_command(command)
        if show:
            print(output)
        if self.status:
            print(f'\n\n *** An error occurred running the command {command}\n\n')
        return self.status
    
    def check_for_intel_gpu(self, pkg_mgr):
        output = self.run_command("lspci -k | grep -EA3 'VGA|3D|Display'")
        if not "Intel" in output:
            return
        print(output)
        response = input("\nIn order to run YOLO on Intel iGPU, some drivers need to be installed. Would you like to install compute drivers for Intel iGPU? N/y ")
        if not response:
            return
        if response.lower() != "y":
            return

        if pkg_mgr == "apt":

            print(self.run_command("sudo apt install curl"))

            if self.rcwec("curl -OL https://github.com/intel/intel-graphics-compiler/releases/download/v2.10.8/intel-igc-opencl-2_2.10.8+18926_amd64.deb", False): return
            print("Downloaded intel-igc-opencl-2_2.10.8+18926_amd64.deb")
            if self.rcwec("curl -OL https://github.com/intel/compute-runtime/releases/download/25.13.33276.16/intel-level-zero-gpu-dbgsym_1.6.33276.16_amd64.ddeb", False): return
            print("Downloaded intel-level-zero-gpu-dbgsym_1.6.33276.16_amd64.ddeb")
            if self.rcwec("curl -OL https://github.com/intel/compute-runtime/releases/download/25.13.33276.16/intel-opencl-icd-dbgsym_25.13.33276.16_amd64.ddeb", False): return
            print("Downloaded intel-opencl-icd-dbgsym_25.13.33276.16_amd64.ddeb")
            if self.rcwec("curl -OL https://github.com/intel/intel-graphics-compiler/releases/download/v2.10.8/intel-igc-core-2_2.10.8+18926_amd64.deb", False): return
            print("Downloaded intel-igc-core-2_2.10.8+18926_amd64.deb")
            if self.rcwec("curl -OL https://github.com/intel/compute-runtime/releases/download/25.13.33276.16/intel-level-zero-gpu_1.6.33276.16_amd64.deb", False): return
            print("Downloaded intel-level-zero-gpu_1.6.33276.16_amd64.deb")
            if self.rcwec("curl -OL https://github.com/intel/compute-runtime/releases/download/25.13.33276.16/intel-opencl-icd_25.13.33276.16_amd64.deb", False): return
            print("Downloaded intel-opencl-icd_25.13.33276.16_amd64.deb")
            if self.rcwec("curl -OL https://github.com/intel/compute-runtime/releases/download/25.13.33276.16/libigdgmm12_22.7.0_amd64.deb", False): return
            print("Downloaded libigdgmm12_22.7.0_amd64.deb")

            if self.rcwec("sudo dpkg -i *deb"): return

            if self.rcwec("rm intel-igc-core-2_2.10.8+18926_amd64.deb"): return
            if self.rcwec("rm intel-igc-opencl-2_2.10.8+18926_amd64.deb"): return
            if self.rcwec("rm intel-level-zero-gpu-dbgsym_1.6.33276.16_amd64.ddeb"): return
            if self.rcwec("rm intel-level-zero-gpu_1.6.33276.16_amd64.deb"): return
            if self.rcwec("rm intel-opencl-icd-dbgsym_25.13.33276.16_amd64.ddeb"): return
            if self.rcwec("rm intel-opencl-icd_25.13.33276.16_amd64.deb"): return
            if self.rcwec("rm libigdgmm12_22.7.0_amd64.deb"): return

        if pkg_mgr == "dnf":

            print(self.run_command("sudo dnf install -y intel-compute-runtime"))

        if pkg_mgr == "pacman":
            output = self.run_command("pacman -Q | grep intel-compute-runtime")
            if output:
                print("Intel compute runtime is already installed on this system")
            else:
                print(self.run_command("sudo pacman -S --noconfirm intel-compute-runtime"))

    def check_for_intel_npu(self, pkg_mgr):
        if not pkg_mgr == "apt":
            return

        print("Checking for Intel NPU ...")
        output = self.run_command("lscpu | grep 'Model name:'")
        print(f"CPU: {output.split(':')[1].strip()}")
        if not "Intel(R) Core(TM) Ultra" in output:
            print("Did not find Intel NPU")
            return

        response = input("Would you like to install optional drivers for Intel NPU? N/y ")
        if not response:
            return
        if response.lower() != "y":
            return

        if self.rcwec("wget https://github.com/oneapi-src/level-zero/releases/download/v1.20.2/level-zero_1.20.2+u22.04_amd64.deb"): return
        if self.rcwec("wget https://github.com/intel/linux-npu-driver/releases/download/v1.16.0/intel-level-zero-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb"): return
        if self.rcwec("wget https://github.com/intel/linux-npu-driver/releases/download/v1.16.0/intel-fw-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb"): return
        if self.rcwec("wget https://github.com/intel/linux-npu-driver/releases/download/v1.16.0/intel-driver-compiler-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb"): return
        if self.rcwec("sudo apt install -y libtbb12"): return
        if self.rcwec("sudo dpkg -i *.deb"): return
        if self.rcwec("sudo chown root:render /dev/accel/accel0"): return
        if self.rcwec("sudo chmod g+rw /dev/accel/accel0"): return
        if self.rcwec("sudo usermod -a -G render $USER"): return
        if self.rcwec("sudo bash -c \"echo 'SUBSYSTEM==\"accel\", KERNEL==\"accel*\", GROUP=\"render\", MODE=\"0660\"' > /etc/udev/rules.d/10-intel-vpu.rules\""): return
        if self.rcwec("sudo udevadm control --reload-rules"): return
        if self.rcwec("sudo udevadm trigger --subsystem-match=accel"): return

        self.rcwec("rm level-zero_1.20.2+u22.04_amd64.deb")
        self.rcwec("rm intel-level-zero-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb")
        self.rcwec("rm intel-fw-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb")
        self.rcwec("rm intel-driver-compiler-npu_1.16.0.20250328-14132024782_ubuntu24.04_amd64.deb")

        if response := input("The system needs to reboot to enable the NPU drivers, would you like to reboot now? N/y  "):
            if response.lower() == "y":
                self.run_command("sudo reboot now")

    def get_host_dist(self):
        result = None
        try:
            with open("/etc/os-release", 'r') as file:
                for line in file:
                    line = line.strip()
                    key, val = line.split("=")
                    if key == "PRETTY_NAME":
                        result = val.strip('"')
        except Exception as ex:
            print(f"File read error /etc/os-release: {ex}")
        return result
    
    def collect_user_sudo(self):
        self.run_command('sudo ls -l')

    def install_icon(self):
        print("\n")
        print(self.run_command(f'sudo {self.venv}/bin/onvif-gui --icon'))
        print("\n")

    def uninstall(self):
        try:
            home = os.environ.get('HOME')
            targets = [f"{home}/.config/onvif-gui", f"{home}/.cache/onvif-gui"]
            icon_filename = "/usr/share/applications/onvif-gui.desktop"
            if not os.path.exists(icon_filename):
                raise Exception(f"Did not find installed onvif gui icon at {icon_filename}")

            with open(icon_filename, 'r') as file:
                for line in file:
                    tokens = line.split("=")
                    if tokens[0] == "Exec":
                        vars = tokens[1].split()
                        install_directory = vars[0][:vars[0].find('/bin/onvif-gui')]
                        targets.append(install_directory)
            
            for target in targets:
                if os.path.exists(target):
                    shutil.rmtree(target)

            output = self.run_command(f'sudo rm {icon_filename}')
            if self.status:
                raise Exception(output)

            print("Onvif GUI has been removed from the system")

        except Exception as ex:
            print(f"An error occurred during un-installation: {ex}")

if __name__ == "__main__":

    if not(home := os.environ.get('HOME')):
        sys.exit(1)

    install = Install()

    parser = argparse.ArgumentParser(description="Install Onvif GUI using an automated script")
    parser.add_argument("-u", "--uninstall", action="store_true", help="Uninstall")

    args = parser.parse_args()

    if args.uninstall:
        install.uninstall()
        sys.exit(0)

    install.collect_user_sudo()

    print(f"Operating System of host: {install.get_host_dist()}")
    
    if not(pkg_mgr := install.find_package_manager()):
        sys.exit(1)

    if install.check_python(pkg_mgr):
        sys.exit(1)

    venv = f"{os.environ.get('HOME')}/.local/share/onvif-gui-env"
    if install.create_venv(venv, pkg_mgr):
        sys.exit(1)
    
    if INSTALL_SOURCE == "pypi":
        if install.install_onvif_gui(venv):
            sys.exit(1)
    elif INSTALL_SOURCE == "smb":
        if install.install_onvif_gui_local(venv, pkg_mgr):
            sys.exit(1)

    install.check_display_protocol(pkg_mgr)
    install.check_for_nvidia()
    install.check_for_intel_gpu(pkg_mgr)

    install.install_icon()

    install.check_for_intel_npu(pkg_mgr)



#*******************************************************************************
# libonvif/setup.py
#
# Copyright (c) 2023, 2024 Stephen Rhodes 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#******************************************************************************/

import os
import sys
import subprocess
import distutils.ccompiler
from pathlib import Path
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

PKG_NAME = "libonvif"
VERSION = "3.2.9"

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        extdir = os.path.join(extdir, PKG_NAME)

        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DCMAKE_POLICY_VERSION_MINIMUM=3.5",
            f"-Wno-dev",]
        
        if sys.platform == "win32":
            cmake_args.append(f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_RELEASE={extdir}")
        
        # add libxml2 install dir to path so native FindLibXml2 succeeds
        directory = os.path.join(Path(__file__).parent.absolute(), "libxml2")
        os.environ['PATH'] = directory + os.pathsep + os.environ['PATH']

        build_temp = os.path.join(self.build_temp, ext.name)
        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        subprocess.run(["cmake", ext.sourcedir] + cmake_args, cwd=build_temp)
        subprocess.run(["cmake", "--build", ".", "--config", "Release"], cwd=build_temp)

def get_package_data():
    data = []
    for f in os.listdir(PKG_NAME):
        if f != "__init__.py":
            data.append(f)
    return data

setup(
    name=PKG_NAME,
    version=VERSION,
    author="Stephen Rhodes",
    author_email="sr99622@gmail.com",
    description="A client side onvif library",
    long_description="",
    ext_modules=[CMakeExtension(PKG_NAME)],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    python_requires=">=3.10",
    packages=[PKG_NAME],
    package_data={ PKG_NAME : get_package_data() }
)
